import argparse
import ast
import itertools
import json
import socket
import sys
import traceback
import typing as T
from datetime import datetime, timezone
from pathlib import Path
from textwrap import dedent

import numpy as np
import optuna
import ray
from llama_index.core import VectorStoreIndex
from llama_index.core.retrievers import BaseRetriever
from llama_index.core.schema import NodeWithScore, TextNode
from openai import APIConnectionError
from ray.util import state

from syftr.baselines import set_baselines
from syftr.evaluation.evaluation import eval_dataset
from syftr.flows import (
    CoAAgentFlow,
    ConsensusJudgeFlow,
    CritiqueAgentFlow,
    Flow,
    JudgeFlow,
    LATSAgentFlow,
    RAGFlow,
    RandomJudgeFlow,
    ReActAgentFlow,
    RetrieverFlow,
    SubQuestionRAGFlow,
)
from syftr.hf_endpoint_embeddings import HFEndpointEmbeddings
from syftr.huggingface_helper import get_embedding_model
from syftr.logger import logger
from syftr.optimization import StudyRunner, get_study
from syftr.optuna_helper import (
    get_pareto_flows,
    run_flows,
    trial_exists,
    without_non_search_space_params,
)
from syftr.output_parsers.judge import (
    parse_correctness_evaluation,
    parse_correctness_evaluation_comparison,
    parse_correctness_evaluation_simple,
    parse_correctness_evaluation_ten,
)
from syftr.prompts.judge import (
    COMPARISON_JUDGE_QUERY_PROMPT_TEMPLATE,
    DEFAULT_JUDGE_QUERY_PROMPT_TEMPLATE,
    DEFAULT_JUDGE_SYSTEM_PROMPT,
    JUDGE_SYSTEM_PROMPT_COMPARISON,
    JUDGE_SYSTEM_PROMPT_DETAILED,
    JUDGE_SYSTEM_PROMPT_SIMPLE,
    JUDGE_SYSTEM_PROMPT_TEN,
)
from syftr.ray.utils import ray_init
from syftr.retrievers.build import build_rag_retriever
from syftr.startup import prepare_worker
from syftr.studies import (
    RetrieverStudyConfig,
    StudyConfig,
    get_default_study_name,
    get_pareto_study_config,
)
from syftr.templates import get_template
from syftr.tracing import get_span_exporter, set_tracing_metrics


def _get_examples(example_retriever: BaseRetriever, query_str: str):
    retrieved_nodes: T.List[NodeWithScore] = example_retriever.retrieve(query_str)
    result_strs = []
    for n in retrieved_nodes:
        try:
            raw_dict = ast.literal_eval(n.text)
            query = raw_dict["query"]
            response = raw_dict["response"]
            result_str = dedent(
                f"""\
                Question: {query}
                Answer: {response}"""
            )
            result_strs.append(result_str)
        except SyntaxError as exc:
            logger.warning("Converting example to dictionary failed: %s", exc)
            result_strs.append(n.text)
    return "\n\n".join(result_strs)


# HERE
def _get_example_retriever(params, study_config: StudyConfig, embedding_model):
    assert embedding_model, "No embedding model for dynamic few-shot prompting"
    logger.info("Building few-shot retriever")
    dataset_iter = study_config.dataset.iter_examples(partition="train")
    logger.info("Getting few-shot examples from dataset")
    if study_config.toy_mode:
        dataset_iter = itertools.islice(dataset_iter, 20)
    few_shot_nodes = []
    for pair in dataset_iter:
        line = f"{{'query': '''{pair.question}''', 'response': '''{pair.answer}'''}}"
        few_shot_nodes.append(TextNode(text=line))

    if not isinstance(embedding_model, HFEndpointEmbeddings):
        embedding_model.reset_timeouts(total_chunks=len(few_shot_nodes))
    logger.info(
        f"Building few-shot retriever index using embedding model {embedding_model}"
    )
    try:
        few_shot_index = VectorStoreIndex(
            nodes=few_shot_nodes, embed_model=embedding_model
        )
    except APIConnectionError:
        logger.error(
            "Failed to build few-shot retriever index. "
            f"This is likely due to a timeout with the {embedding_model} embedding model."
        )
        raise
    logger.info(
        f"Built few-shot retriever index using embedding model {embedding_model}"
    )
    few_shot_retriever = few_shot_index.as_retriever(
        similarity_top_k=params["few_shot_top_k"], similarity_threshold=None
    )

    def get_qa_examples(query_str, **kwargs):
        return _get_examples(few_shot_retriever, query_str)

    return get_qa_examples


def evaluate(
    params: T.Dict,
    study_config: StudyConfig,
) -> T.Tuple[float, float, T.Dict[str, T.Any], str]:
    prepare_worker()
    flow_start = datetime.now(timezone.utc).timestamp()
    logger.info("Evaluating flow with config: %s", params)
    flow_json = json.dumps(params)
    if study_config.evaluation.use_tracing_metrics:
        span_exporter = get_span_exporter()
    obj1, obj2, results = _evaluate(params, study_config)
    if study_config.evaluation.use_tracing_metrics:
        set_tracing_metrics(span_exporter, results)
    results["failed"] = False
    results["flow_start"] = flow_start
    results["flow_end"] = datetime.now(timezone.utc).timestamp()
    results["flow_duration"] = float(results["flow_end"]) - float(results["flow_start"])
    logger.info("Evaluation finished. Finalizing trial report. %s", results)
    return obj1, obj2, results, flow_json


def build_flow(params: T.Dict, study_config: StudyConfig) -> Flow:
    from syftr.llm import get_llm

    enforce_full_evaluation = params.get("enforce_full_evaluation", False)
    use_reasoning = params.get("use_reasoning")

    if study_config.is_judge_study:
        prompt = DEFAULT_JUDGE_SYSTEM_PROMPT
        output_parser = parse_correctness_evaluation
        template = DEFAULT_JUDGE_QUERY_PROMPT_TEMPLATE
        if params["judge_prompt"] == "simple":
            prompt = JUDGE_SYSTEM_PROMPT_SIMPLE
            output_parser = parse_correctness_evaluation_simple  # type: ignore
        elif params["judge_prompt"] == "out_of_ten":
            prompt = JUDGE_SYSTEM_PROMPT_TEN
            output_parser = parse_correctness_evaluation_ten  # type: ignore
        elif params["judge_prompt"] == "detailed":
            prompt = JUDGE_SYSTEM_PROMPT_DETAILED
            output_parser = parse_correctness_evaluation
        elif params["judge_prompt"] == "comparison":
            prompt = JUDGE_SYSTEM_PROMPT_COMPARISON
            output_parser = parse_correctness_evaluation_comparison  # type: ignore
            template = COMPARISON_JUDGE_QUERY_PROMPT_TEMPLATE

        if params["judge_type"] == "single_correctness_evaluator":
            response_synthesizer_llm = get_llm(params["response_synthesizer_llm_name"])
            return JudgeFlow(
                response_synthesizer_llm=response_synthesizer_llm,
                system_prompt=prompt,
                output_parser=output_parser,
                query_prompt_template=template,
                params=params,
                enforce_full_evaluation=enforce_full_evaluation,
                temperature=params["response_synthesizer_llm_temperature"],
            )

        llm_names = ast.literal_eval(params["response_synthesizer_llm_combination"])
        response_synthesizer_llms = [get_llm(llm_name) for llm_name in llm_names]

        if params["judge_type"] == "random_correctness_evaluator":
            return RandomJudgeFlow(
                response_synthesizer_llms=response_synthesizer_llms,  # type: ignore
                system_prompt=prompt,
                output_parser=output_parser,
                query_prompt_template=template,
                params=params,
                enforce_full_evaluation=enforce_full_evaluation,
                temperature=params["response_synthesizer_llm_temperature"],
            )

        if params["judge_type"] == "consensus_correctness_evaluator":
            return ConsensusJudgeFlow(
                response_synthesizer_llms=response_synthesizer_llms,  # type: ignore
                system_prompt=prompt,
                output_parser=output_parser,
                query_prompt_template=template,
                params=params,
                enforce_full_evaluation=enforce_full_evaluation,
                temperature=params["response_synthesizer_llm_temperature"],
            )

        raise RuntimeError("Unknown judge type: %s" % params["judge_type"])

    response_synthesizer_llm = get_llm(
        name=params["response_synthesizer_llm_name"],
        temperature=params.get("response_synthesizer_temperature"),
        top_p=params.get("response_synthesizer_top_p"),
    )

    if study_config.is_retriever_study:
        hyde_llm = (
            get_llm(
                name=params["hyde_llm_name"],
                temperature=params.get("hyde_llm_temperature"),
                top_p=params.get("hyde_llm_top_p"),
            )
            if params.get("hyde_enabled")
            else None
        )
        retriever, docstore = build_rag_retriever(study_config, params)
        return RetrieverFlow(
            response_synthesizer_llm=response_synthesizer_llm,
            retriever=retriever,
            docstore=docstore,
            hyde_llm=hyde_llm,
            additional_context_num_nodes=params.get("additional_context_num_nodes", 0),
            params=params,
            enforce_full_evaluation=enforce_full_evaluation,
            use_reasoning=use_reasoning,
        )

    get_qa_examples = None
    is_few_shot = study_config.search_space.is_few_shot(params)
    if is_few_shot:
        few_shot_embedding_model_name = params["few_shot_embedding_model"]
        few_shot_embedding_model, _ = get_embedding_model(
            few_shot_embedding_model_name,
            timeout_config=study_config.timeouts,
            device=study_config.optimization.embedding_device,
            use_hf_endpoint_models=study_config.optimization.use_hf_embedding_models,
        )
        get_qa_examples = _get_example_retriever(
            params, study_config, few_shot_embedding_model
        )

    do_rag = params["rag_mode"] != "no_rag"
    template_name = params["template_name"]
    template = get_template(
        template_name, with_context=do_rag, with_few_shot_prompt=is_few_shot
    )

    flow: T.Any

    if not do_rag:
        flow = Flow(
            response_synthesizer_llm=response_synthesizer_llm,
            template=template,
            get_examples=get_qa_examples,
            params=params,
            enforce_full_evaluation=enforce_full_evaluation,
            use_reasoning=use_reasoning,
        )
    else:
        hyde_llm = reranker_llm = reranker_top_k = None
        if params.get("hyde_enabled"):
            hyde_llm = get_llm(
                name=params["hyde_llm_name"],
                temperature=params.get("hyde_llm_temperature"),
                top_p=params.get("hyde_llm_top_p"),
            )
        if params.get("reranker_enabled"):
            reranker_llm = get_llm(
                name=params["reranker_llm_name"],
                temperature=params.get("reranker_llm_temperature"),
                top_p=params.get("reranker_llm_top_p"),
            )
            reranker_top_k = params["reranker_top_k"]
        if params.get("additional_context_enabled"):
            additional_context_num_nodes = params["additional_context_num_nodes"]
        else:
            additional_context_num_nodes = 0

        rag_retriever, rag_docstore = build_rag_retriever(study_config, params)

        match params["rag_mode"]:
            case "rag":
                flow = RAGFlow(
                    retriever=rag_retriever,
                    response_synthesizer_llm=response_synthesizer_llm,
                    docstore=rag_docstore,
                    template=template,
                    get_examples=get_qa_examples,
                    hyde_llm=hyde_llm,
                    reranker_llm=reranker_llm,
                    reranker_top_k=reranker_top_k,
                    additional_context_num_nodes=additional_context_num_nodes,
                    enforce_full_evaluation=enforce_full_evaluation,
                    use_reasoning=use_reasoning,
                    params=params,
                )
            case "react_rag_agent":
                subquestion_engine_llm = get_llm(
                    name=params["subquestion_engine_llm_name"],
                    temperature=params.get("subquestion_engine_llm_temperature"),
                    top_p=params.get("subquestion_engine_llm_top_p"),
                )
                subquestion_response_synthesizer_llm = get_llm(
                    name=params["subquestion_response_synthesizer_llm_name"],
                    temperature=params.get(
                        "subquestion_response_synthesizer_llm_temperature"
                    ),
                    top_p=params.get("subquestion_response_synthesizer_llm_top_p"),
                )
                flow = ReActAgentFlow(
                    retriever=rag_retriever,
                    response_synthesizer_llm=response_synthesizer_llm,
                    subquestion_response_synthesizer_llm=subquestion_response_synthesizer_llm,
                    subquestion_engine_llm=subquestion_engine_llm,
                    docstore=rag_docstore,
                    template=template,
                    get_examples=get_qa_examples,
                    hyde_llm=hyde_llm,
                    reranker_llm=reranker_llm,
                    reranker_top_k=reranker_top_k,
                    additional_context_num_nodes=additional_context_num_nodes,
                    dataset_name=study_config.dataset.name,
                    dataset_description=study_config.dataset.description,
                    enforce_full_evaluation=enforce_full_evaluation,
                    use_reasoning=use_reasoning,
                    params=params,
                )
            case "critique_rag_agent":
                subquestion_engine_llm = get_llm(
                    name=params["subquestion_engine_llm_name"],
                    temperature=params.get("subquestion_engine_llm_temperature"),
                    top_p=params.get("subquestion_engine_llm_top_p"),
                )
                subquestion_response_synthesizer_llm = get_llm(
                    name=params["subquestion_response_synthesizer_llm_name"],
                    temperature=params.get(
                        "subquestion_response_synthesizer_llm_temperature"
                    ),
                    top_p=params.get("subquestion_response_synthesizer_llm_top_p"),
                )
                critique_agent_llm = get_llm(
                    name=params["critique_agent_llm_name"],
                    temperature=params.get("critique_agent_llm_temperature"),
                    top_p=params.get("critique_agent_llm_top_p"),
                )
                reflection_agent_llm = get_llm(
                    name=params["reflection_agent_llm_name"],
                    temperature=params.get("reflection_agent_llm_temperature"),
                    top_p=params.get("reflection_agent_llm_top_p"),
                )
                flow = CritiqueAgentFlow(
                    response_synthesizer_llm=response_synthesizer_llm,
                    subquestion_engine_llm=subquestion_engine_llm,
                    subquestion_response_synthesizer_llm=subquestion_response_synthesizer_llm,
                    critique_agent_llm=critique_agent_llm,
                    reflection_agent_llm=reflection_agent_llm,
                    retriever=rag_retriever,
                    docstore=rag_docstore,
                    template=template,
                    get_examples=get_qa_examples,
                    hyde_llm=hyde_llm,
                    reranker_llm=reranker_llm,
                    reranker_top_k=reranker_top_k,
                    additional_context_num_nodes=additional_context_num_nodes,
                    dataset_name=study_config.dataset.name,
                    dataset_description=study_config.dataset.description,
                    enforce_full_evaluation=enforce_full_evaluation,
                    use_reasoning=use_reasoning,
                    params=params,
                )
            case "sub_question_rag":
                subquestion_engine_llm = get_llm(
                    name=params["subquestion_engine_llm_name"],
                    temperature=params.get("subquestion_engine_llm_temperature"),
                    top_p=params.get("subquestion_engine_llm_top_p"),
                )
                subquestion_response_synthesizer_llm = get_llm(
                    name=params["subquestion_response_synthesizer_llm_name"],
                    temperature=params.get(
                        "subquestion_response_synthesizer_llm_temperature"
                    ),
                    top_p=params.get("subquestion_response_synthesizer_llm_top_p"),
                )
                flow = SubQuestionRAGFlow(
                    response_synthesizer_llm=response_synthesizer_llm,
                    subquestion_engine_llm=subquestion_engine_llm,
                    subquestion_response_synthesizer_llm=subquestion_response_synthesizer_llm,
                    retriever=rag_retriever,
                    docstore=rag_docstore,
                    template=template,
                    get_examples=get_qa_examples,
                    hyde_llm=hyde_llm,
                    reranker_llm=reranker_llm,
                    reranker_top_k=reranker_top_k,
                    additional_context_num_nodes=additional_context_num_nodes,
                    dataset_name=study_config.dataset.name,
                    dataset_description=study_config.dataset.description,
                    enforce_full_evaluation=enforce_full_evaluation,
                    use_reasoning=use_reasoning,
                    params=params,
                )
            case "lats_rag_agent":
                flow = LATSAgentFlow(
                    retriever=rag_retriever,
                    response_synthesizer_llm=response_synthesizer_llm,
                    docstore=rag_docstore,
                    template=template,
                    get_examples=get_qa_examples,
                    hyde_llm=hyde_llm,
                    reranker_llm=reranker_llm,
                    reranker_top_k=reranker_top_k,
                    additional_context_num_nodes=additional_context_num_nodes,
                    dataset_name=study_config.dataset.name,
                    dataset_description=study_config.dataset.description,
                    num_expansions=params["lats_num_expansions"],
                    max_rollouts=params["lats_max_rollouts"],
                    enforce_full_evaluation=enforce_full_evaluation,
                    use_reasoning=use_reasoning,
                    params=params,
                )
            case "coa_rag_agent":
                flow = CoAAgentFlow(
                    retriever=rag_retriever,
                    response_synthesizer_llm=response_synthesizer_llm,
                    docstore=rag_docstore,
                    template=template,
                    get_examples=get_qa_examples,
                    hyde_llm=hyde_llm,
                    reranker_llm=reranker_llm,
                    reranker_top_k=reranker_top_k,
                    additional_context_num_nodes=additional_context_num_nodes,
                    dataset_name=study_config.dataset.name,
                    dataset_description=study_config.dataset.description,
                    enable_calculator=params["coa_enable_calculator"],
                    enforce_full_evaluation=enforce_full_evaluation,
                    use_reasoning=use_reasoning,
                    params=params,
                )
            case _:
                raise ValueError(f"Invalid rag_mode: {params['rag_mode']}")

    return flow


def _evaluate(
    params: T.Dict,
    study_config: StudyConfig,
) -> T.Tuple[float, float, T.Dict[str, float | str]]:
    flow = build_flow(params, study_config)

    results: T.Dict[str, T.Any] = eval_dataset(
        study_config=study_config,
        dataset_iter=study_config.dataset,
        flow=flow,
        evaluation_mode=study_config.evaluation.mode,
    )

    obj1 = results[study_config.optimization.objective_1_name]
    obj2 = results[study_config.optimization.objective_2_name]
    if np.isnan(obj2):
        logger.fatal(
            "%s value is NaN and the trial will crash.\nParams: %s",
            study_config.optimization.objective_2_name,
            json.dumps(params, indent=2),
        )
    return obj1, obj2, results


def objective(
    trial: optuna.Trial,
    study_config: StudyConfig,
    components: T.List[str],
) -> T.Tuple[float, float]:
    from syftr.tuner.core import set_trial

    logger.debug("Syftr objective is running with: %s", sys.executable)

    search_space = study_config.search_space
    params: dict[str, str | bool | int | float]
    for i in range(study_config.optimization.num_retries_unique_params):
        params = search_space.sample(trial, components)  # type: ignore
        if not study_config.optimization.skip_existing:
            logger.info("Using generated parameter combination without check")
            break
        if not trial_exists(study_config.name, params):
            logger.info(
                "Found novel parameter combination after %i retries: %s",
                i,
                str(params),
            )
            break
    try:
        obj1, obj2, metrics, flow_json = evaluate(params, study_config)
    except Exception as ex:
        logger.exception("Objective had an unhandled exception: %s", ex)
        metrics = {
            "failed": True,
            "exception_message": str(ex),
            "exception_stacktrace": traceback.format_exc(),
            "exception_class": ex.__class__.__name__,
        }
        flow_json = json.dumps(params)
        raise ex
    finally:
        set_trial(
            trial=trial,
            study_config=study_config,
            params=params,
            is_seeding=False,
            metrics=metrics,
            flow_json=flow_json,
        )
        worker_state = state.get_worker(ray.get_runtime_context().get_worker_id())
        if worker_state is not None:
            trial.set_user_attr("ray_worker_pid", worker_state.pid)  # type: ignore
        trial.set_user_attr("ray_worker_hostname", socket.gethostname())
    logger.debug("Optimizer finished trial with params: %s", trial.params)
    return obj1, obj2


@ray.remote
def run_flow(flows: T.Dict[str, T.Any], study_config: StudyConfig) -> None:
    try:
        _run_flow(flows, study_config)
    except Exception:
        logger.exception("Trial had an unhandled exception: %s", flows)


def _run_flow(flow: T.Dict[str, T.Any], study_config: StudyConfig) -> None:
    from syftr.configuration import cfg
    from syftr.tuner.core import set_trial

    logger.debug("Syftr seeder is running with: %s", sys.executable)

    study_name = study_config.name
    params = without_non_search_space_params(flow, study_config)
    search_space = study_config.search_space

    logger.info("Loading study: %s", study_name)
    study = optuna.load_study(
        study_name=study_name, storage=cfg.database.get_optuna_storage()
    )

    if study_config.optimization.skip_existing and trial_exists(
        study_name, params, cfg.database.get_optuna_storage()
    ):
        logger.warning(
            "Flow already exists in study '%s': %s",
            study_name,
            str(flow),
        )
        return

    logger.info("Seeding study '%s' with baseline: %s", study_name, str(params))
    distributions = search_space.build_distributions(params=params)
    try:
        obj1, obj2, metrics, flow_json = evaluate(flow, study_config)
        trial = optuna.create_trial(
            values=[obj1, obj2],
            params=params,
            distributions=distributions,
        )
    except Exception as ex:
        logger.exception("Seeding had an unhandled exception: %s", ex)
        trial = optuna.create_trial(
            params=params,
            distributions=distributions,
            state=optuna.trial.TrialState.FAIL,
        )
        metrics = {
            "failed": True,
            "exception_message": str(ex),
            "exception_stacktrace": traceback.format_exc(),
            "exception_class": ex.__class__.__name__,
        }
        flow_json = json.dumps(params)
        raise ex
    set_trial(trial, study_config, params, True, metrics, flow_json)
    worker_state = state.get_worker(ray.get_runtime_context().get_worker_id())
    if worker_state is not None:
        trial.set_user_attr("ray_worker_pid", worker_state.pid)  # type: ignore
    trial.set_user_attr("ray_worker_hostname", socket.gethostname())
    study.add_trial(trial)
    logger.debug("Seeding added trial with params: %s", trial.params)


def run_pareto_flows(study_config: StudyConfig):
    assert study_config.pareto is not None, "Pareto config is not set"
    if study_config.pareto.raise_on_same_study:
        assert study_config.pareto.name != study_config.name, (
            "Pareto study is the same as the main study"
        )
    pareto_flows = get_pareto_flows(study_config)
    logger.info(
        "Evaluating %i Pareto-flows from %s using study %s",
        len(pareto_flows),
        study_config.name,
        study_config.pareto.name,
    )
    pareto_study_config = get_pareto_study_config(study_config)

    # if we want to replace llm name with a single llm name
    if study_config.pareto.replacement_llm_name:
        for flow in pareto_flows:
            study_config.replace_llm_name(flow)

    get_study(pareto_study_config)
    logger.info("Pareto study config is: %s", pareto_study_config)
    run_flows(
        flows=pareto_flows,
        study_config=pareto_study_config,
        runner=run_flow,
    )


def run(
    study_config: StudyConfig | str,
    skip_optimization: bool = False,
    skip_pareto: bool = True,
    remote: bool = False,
) -> None:
    if isinstance(study_config, str):
        study_config_file = Path(study_config)
        study_config = StudyConfig.from_file(study_config_file)
        if study_config.evaluation.mode == "retriever":
            study_config = RetrieverStudyConfig.from_file(study_config_file)
    logger.info("Active configuration is: %s", study_config)

    ray_init(remote)

    if skip_optimization:
        logger.info("Skipping optimization")
    else:
        logger.info("Running optimization using study: %s", study_config.name)
        study_config = set_baselines(study_config)
        # Kick off the study
        optimization = StudyRunner(
            objective=objective,
            study_config=study_config,
            seeder=run_flow,
        )
        optimization.run()

    # Evaluate the Pareto front as configured
    if skip_pareto:
        logger.info("Skipping Pareto evaluation")
    else:
        if study_config.pareto:
            run_pareto_flows(study_config)
        else:
            logger.warning("No Pareto config found in study config")


def main():
    parser = argparse.ArgumentParser()
    default = get_default_study_name()
    parser.add_argument(
        "--study-config",
        help=f"Path to study config yaml (default: {default})",
        default=default,
    )
    parser.add_argument(
        "--skip-optimization",
        help="Do not run the optimization if set",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--skip-pareto",
        help="Do not run the the separate Pareto evaluation",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--remote",
        help="Run the optimization remotely",
        default=False,
        action="store_true",
    )
    args = parser.parse_args()
    run(
        study_config=args.study_config,
        skip_optimization=args.skip_optimization,
        skip_pareto=args.skip_pareto,
        remote=args.remote,
    )


if __name__ == "__main__":
    main()



import hashlib
import io
import json
from contextlib import contextmanager
from pickle import UnpicklingError
from typing import Any, Optional

import cloudpickle
import diskcache
import torch
from llama_index.core.indices.base import BaseIndex
from lz4.frame import compress, decompress

from syftr.amazon import delete_file_from_s3, get_file_from_s3
from syftr.configuration import cfg
from syftr.logger import logger
from syftr.studies import StudyConfig
from syftr.utils.locks import distributed_lock

original_torch_load = torch.load


def patched_torch_load(*args, **kwargs):
    kwargs["map_location"] = torch.device("cpu")
    return original_torch_load(*args, **kwargs)


torch.load = patched_torch_load


@contextmanager
def local_cache():
    with diskcache.Cache(
        cfg.paths.index_cache, size_limit=cfg.storage.local_cache_max_size_gb * 1024**3
    ) as cache:
        yield cache


def _get_dense_index_params_hash(study_config, params, cache_version: int = 3) -> str:
    param_names = [
        "splitter_chunk_overlap_frac",
        "splitter_chunk_exp",
        "splitter_method",
        "rag_embedding_model",
    ]
    index_params = {key: value for key, value in params.items() if key in param_names}
    dataset_param_names = ["xname", "partition_map", "subset", "grounding_data_path"]
    dataset_params = {
        key: value
        for key, value in study_config.dataset.model_dump().items()
        if key in dataset_param_names
    }
    key_params = {
        "index": index_params,
        "dataset": dataset_params,
        "cache_version": cache_version,
    }
    params_hash = hashlib.sha1(
        json.dumps(key_params, sort_keys=True).encode("utf-8")
    ).hexdigest()
    logger.info(
        f"Built cache key {params_hash} for params {json.dumps(key_params, sort_keys=True)}"
    )
    return params_hash


@contextmanager
def index_cache_lock(study_config: StudyConfig, params):
    cache_key = _get_dense_index_params_hash(study_config, params)
    host_only = not cfg.storage.s3_cache_enabled
    with distributed_lock(cache_key, host_only=host_only):
        yield cache_key


def put_cache(cache_key, index, local_only: bool = False) -> None:
    serialized_obj = compress(cloudpickle.dumps(index))
    try:
        index = cloudpickle.loads(decompress(serialized_obj))
        assert isinstance(index, BaseIndex), index
    except (UnpicklingError, AssertionError) as exc:
        logger.warning(
            "Pickling index failed with exception %s. Not storing object.", exc
        )
        return

    with local_cache() as cache:
        logger.info(f"Storing index to {cache.directory}")
        cache.add(cache_key, serialized_obj)
        logger.info(f"Done storing index to {cache.directory}")

    if not local_only and cfg.storage.s3_cache_enabled:
        try:
            import boto3
            from boto3.s3.transfer import TransferConfig
        except ImportError:
            logger.info("Skipping S3 cache - install boto3 to cache objects in S3")
            return
        s3 = boto3.client("s3")
        config = TransferConfig(multipart_threshold=5 * 1024**3)  # 5GB default limit
        fileobj = io.BytesIO(initial_bytes=serialized_obj)
        s3_cache_key = f"index_cache/{cache_key}.pkl"
        logger.info(f"Storing index to S3: {s3_cache_key}")
        s3.upload_fileobj(
            fileobj, cfg.storage.cache_bucket, s3_cache_key, Config=config
        )
        logger.info(f"Done storing index to S3: {s3_cache_key}")


def get_cached(cache_key: str) -> Optional[Any]:
    s3_cache_key = f"index_cache/{cache_key}.pkl"

    try:
        with local_cache() as cache:
            if (data := cache.get(cache_key)) is not None:
                logger.info(f"Loading pre-built index from {cache.directory}")
                index = cloudpickle.loads(decompress(data))
                logger.info(f"Loaded pre-built index from {cache.directory}")
                return index

        if cfg.storage.s3_cache_enabled:
            if (data := get_file_from_s3(s3_cache_key)) is not None:
                logger.info(f"Loading pre-built index from S3: {s3_cache_key}")
                index = cloudpickle.loads(decompress(data))
                assert isinstance(index, BaseIndex), index
                logger.info(f"Loaded pre-built index from S3: {s3_cache_key}")
                put_cache(cache_key, index, local_only=True)
                return index
    except UnpicklingError as exc:
        logger.warning(
            "Unpickling index failed with exception %s. Removing object from caches.",
            exc,
        )
        try:
            del cache[cache_key]
        except Exception:
            pass
        try:
            delete_file_from_s3(s3_cache_key)
        except Exception:
            pass
        return None
    return None



import io
import json

import yaml

from syftr.configuration import cfg


def test_model_dump_json():
    """Test that all fields can be serialized to JSON."""
    data = cfg.json()
    json.loads(data)


def test_model_dump_yaml():
    """Test that all fields can be serialized to YAML."""
    buf = io.StringIO()
    data = json.loads(cfg.model_dump_json())
    yaml.safe_dump(data, buf)
    buf.seek(0)
    yaml.safe_load(buf)



import asyncio
import json
from pathlib import Path
from typing import List

import click
import pandas as pd
from aiolimiter import AsyncLimiter
from llama_index.core.evaluation import CorrectnessEvaluator

from syftr.core import RandomTrial
from syftr.evaluation.evaluation import SyftrEvaluationResult, _aeval_pair
from syftr.instrumentation.arize import instrument_arize
from syftr.llm import get_llm
from syftr.studies import StudyConfig
from syftr.tuner.qa_tuner import build_flow


@click.command()
@click.option(
    "-s",
    "--study-config-path",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    required=True,
    help="Path to the study config yaml file",
)
@click.option("-i", "--interactive", is_flag=True)
@click.option("-o", "--instrumentation", is_flag=True)
@click.option("-f", "--flow-json", type=str, help="JSON string of flow parameters")
def main(
    study_config_path: Path, flow_json: str, instrumentation: bool, interactive: bool
):
    study_config: StudyConfig = StudyConfig.from_file(study_config_path)
    click.secho(
        f"Loaded study config for study {study_config.name}", bold=True, fg="green"
    )

    if instrumentation:
        instrument_arize()

    if flow_json:
        click.secho(f"Building flow from JSON '{flow_json}'", fg="yellow")
        flow_params = json.loads(flow_json)
    else:
        search_space = study_config.search_space
        random_trial = RandomTrial()
        flow_params = search_space.sample(random_trial)
        flow_json = json.dumps(flow_params, ensure_ascii=False)
        click.secho(f"Building random flow '{flow_json}'", fg="yellow")

    flow = build_flow(flow_params, study_config)
    click.secho(f"Built flow '{flow_json}'", fg="green")

    click.secho(f"Loading QA Pairs from {study_config.dataset.name}", fg="yellow")
    dataset = list(study_config.dataset.iter_examples())
    click.secho(f"Loaded {len(dataset)} QA Pairs", fg="green")

    rate_limiter = AsyncLimiter(
        study_config.optimization.rate_limiter_max_coros,
        study_config.optimization.rate_limiter_period,
    )
    llm_name = study_config.evaluation.llm_names[0]
    llm = get_llm(llm_name)
    evaluator = CorrectnessEvaluator(llm=llm)

    results = []
    for i, pair in enumerate(dataset):
        click.secho("Question: ", fg="blue", nl=False)
        click.echo(f"{pair.question}")
        click.secho("Reference Answer: ", fg="blue", nl=False)
        click.echo(f"{pair.answer}\n")
        if interactive:
            breakpoint()
        result = asyncio.run(
            _aeval_pair(
                pair,
                flow,
                [evaluator],
                rate_limiter,
                raise_on_exception=study_config.evaluation.raise_on_exception,
            )
        )
        passing = result.passing
        if result.generation_exception:
            click.secho(
                f"Generation Exception: {result.generation_exception}", fg="red"
            )
        if result.evaluation_exception:
            click.secho(
                f"Evaluation Exception: {result.evaluation_exception}", fg="red"
            )

        click.echo(f"Generated answer: {result.response}")
        click.secho(
            f"Evaluation Result: {'PASS' if passing else 'FAIL'}: {result.score}/5",
            fg="green" if passing else "red",
        )
        click.secho(f"Evaluation Feedback: {result.feedback}", fg="blue")
        results.append(result)
        df = write_results(results)
        if len(df[~pd.isna(df.passing)]) > 0:
            accuracy = len(df[df.passing]) / len(df)
        else:
            accuracy = 0
        click.echo(f"Current accuracy: {accuracy} ({i + 1}/{len(dataset)} QA pairs)")
        click.echo("=" * 20)
        click.echo("\n")


def write_results(results: List[SyftrEvaluationResult]):
    data = [result.model_dump() for result in results]
    df = pd.DataFrame(data)
    df.to_csv("flow-results.csv")
    return df


if __name__ == "__main__":
    main()



import argparse
import ast
import itertools
import json
import typing as T
from copy import deepcopy
from datetime import datetime, timezone
from pathlib import Path
from textwrap import dedent

import optuna
import ray
from llama_index.core import PromptTemplate, VectorStoreIndex
from llama_index.core.base.base_retriever import BaseRetriever
from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.core.ingestion import IngestionPipeline
from llama_index.core.schema import NodeWithScore, TextNode

from syftr.agent_flows import FlowJSONHandler, LlamaIndexReactRAGAgentFlow
from syftr.baselines import set_baselines
from syftr.evaluation.evaluation import eval_dataset
from syftr.flows import Flow
from syftr.huggingface_helper import get_embedding_model
from syftr.logger import logger
from syftr.optimization import StudyRunner
from syftr.optuna_helper import trial_exists
from syftr.ray.utils import ray_init
from syftr.storage import SyftrQADataset
from syftr.studies import AgentStudyConfig, get_default_study_name
from syftr.templates import get_agent_template, get_template
from syftr.tuner.core import build_splitter


def roundrobin(*iterables):
    """Yields an item from each iterable, alternating between them.

        >>> list(roundrobin('ABC', 'D', 'EF'))
        ['A', 'D', 'E', 'B', 'F', 'C']

    This function produces the same output as :func:`interleave_longest`, but
    may perform better for some inputs (in particular when the number of
    iterables is small).

    """
    pending = len(iterables)
    nexts = itertools.cycle(iter(it).__next__ for it in iterables)
    while pending:
        try:
            for n in nexts:
                yield n()
        except StopIteration:
            pending -= 1
            nexts = itertools.cycle(itertools.islice(nexts, pending))


def _get_examples(example_retriever: BaseRetriever, query_str: str):
    retrieved_nodes: T.List[NodeWithScore] = example_retriever.retrieve(query_str)
    result_strs = []
    for n in retrieved_nodes:
        raw_dict = ast.literal_eval(n.text)
        query = raw_dict["query"]
        response = raw_dict["response"]
        result_str = dedent(
            f"""\
            Question: {query}
            Answer: {response}"""
        )
        result_strs.append(result_str)
    return "\n\n".join(result_strs)


def _build_dense_index(
    dataset: SyftrQADataset,
    transforms,
    embedding_model: BaseEmbedding,
) -> VectorStoreIndex:
    from syftr.configuration import cfg

    pipeline = IngestionPipeline(transformations=transforms)
    nodes = pipeline.run(
        documents=list(dataset.iter_grounding_data()),
        show_progress=cfg.optuna.show_progress,
    )
    return VectorStoreIndex(
        nodes=nodes,
        embed_model=embedding_model,
        insert_batch_size=2048,
        show_progress=cfg.optuna.show_progress,
    )


def _get_example_retriever(params, study_config: AgentStudyConfig):
    if params["prompt_name"] != "few-shot":
        return None, None

    dataset_iter = roundrobin(
        *[ds.iter_grounding_data() for ds in study_config.datasets]
    )
    if study_config.toy_mode:
        dataset_iter = list(itertools.islice(dataset_iter, 20))
    few_shot_nodes = []
    for pair in dataset_iter:
        line = f"{{'query': '''{pair.question}''', 'response': '''{pair.answer}'''}}"
        few_shot_nodes.append(TextNode(text=line))
    embedding_model_name = params.get("embedding_model", None)
    embedding_model, is_onnx = get_embedding_model(embedding_model_name)
    few_shot_index = VectorStoreIndex(nodes=few_shot_nodes, embed_model=embedding_model)
    few_shot_retriever = few_shot_index.as_retriever(
        similarity_top_k=10, similarity_threshold=None
    )

    def get_qa_examples(query_str, **kwargs):
        return _get_examples(few_shot_retriever, query_str)

    return (
        get_qa_examples,
        embedding_model,
    )


def _get_kwargs(params: T.Dict):
    from syftr.llm import get_llm

    kwargs = deepcopy(params)
    kwargs.pop("retriever", None)
    kwargs.pop("embedding_model", None)
    kwargs.pop("splitter", None)
    kwargs.pop("prompt_name", None)
    kwargs.pop("rag_mode", None)
    kwargs["llm_name"] = kwargs.pop("llm")
    reranker_llm = get_llm(kwargs.pop("reranker_llm_name", None))
    if reranker_llm:
        kwargs["reranker_llm_name"] = reranker_llm
    hyde_llm = get_llm(kwargs.pop("hyde_llm_name", None))
    if hyde_llm:
        kwargs["hyde_llm_name"] = hyde_llm
    return kwargs


def evaluate(
    params: T.Dict,
    study_config: AgentStudyConfig,
) -> T.Tuple[float, float, T.Dict[str, T.Any], str]:
    flow_start = datetime.now(timezone.utc).timestamp()
    logger.info("Evaluating agent flow with config: %s", params)
    flow_json = json.dumps(params)
    num_cpus = study_config.optimization.cpus_per_trial
    num_gpus = study_config.optimization.gpus_per_trial

    future = _evaluate.options(num_cpus=num_cpus, num_gpus=num_gpus).remote(
        params, study_config
    )
    timeout_secs = study_config.timeouts.eval_timeout
    logger.info("Launched evaluation task for trial. Trial timeout is %s", timeout_secs)
    try:
        acc, lat, results, flow_json = ray.get(future, timeout=timeout_secs)
    except Exception as exc:
        ray.cancel(future, force=True)
        logger.exception("Agent flow evaluation failed. Reporting error data")
        results = {
            "failed": True,
            "exception": str(exc),
            "exception_class": exc.__class__.__name__,
        }
        raise
    results["flow_start"] = flow_start
    results["flow_end"] = datetime.now(timezone.utc).timestamp()
    results["flow_duration"] = float(results["flow_end"]) - float(results["flow_start"])
    logger.info("Agent flow evaluation finished. Finalizing trial report. %s", results)
    return acc, lat, results, flow_json


@ray.remote
def _evaluate(
    params: T.Dict,
    study_config: AgentStudyConfig,
) -> T.Tuple[float, float, T.Dict[str, float | str], str]:
    from syftr.llm import get_llm

    logger.info("Evaluating flow with config: %s", params)
    params = deepcopy(params)
    response_synthesizer_llm = get_llm(params["llm"])
    get_qa_examples = _get_example_retriever(params, study_config)
    system_template = get_agent_template(params["prompt_name"])

    # For baseline.
    flow: T.Any
    if params["rag_mode"] == "no_rag":
        template = get_template("default")
        flow = Flow(
            response_synthesizer_llm=response_synthesizer_llm,
            template=template,
            get_examples=get_qa_examples,
        )
    elif params["rag_mode"] == "rag":
        embedding_model_name: str = params["embedding_model"]
        embedding_model, is_onnx = get_embedding_model(embedding_model_name)
        template = get_template("default", with_context=True)
        # AgentStudyConfig should be refactored
        splitter = build_splitter(study_config, params)  # type: ignore
        indexes = []
        for ds in study_config.datasets:
            index = _build_dense_index(ds.xname, [splitter], embedding_model)  # type: ignore
            indexes.append((ds.xname, ds.description, index))

        flow = LlamaIndexReactRAGAgentFlow(
            indexes=indexes,
            llm=response_synthesizer_llm,
            template=template,
            system_prompt=PromptTemplate(system_template),
            verbose=True,
            use_hyde=False,
            use_reranker=False,
        )
    else:
        raise ValueError(f"Invalid rag_mode: {params['rag_mode']}")

    flow_json = json.dumps(flow, cls=FlowJSONHandler)

    dataset_iter = roundrobin(
        *[ds.iter_grounding_data() for ds in study_config.datasets]
    )

    results: T.Dict[str, T.Any] = eval_dataset(
        study_config=study_config,
        dataset_iter=dataset_iter,
        flow=flow,
    )
    acc = results["accuracy"]
    lat = results["p80_time"]
    results["is_onnx"] = bool(is_onnx)
    return acc, lat, results, flow_json


def objective(
    trial: optuna.Trial,
    study_config: AgentStudyConfig,
) -> T.Tuple[float, float]:
    from syftr.tuner.core import set_trial

    search_space = study_config.search_space
    params: dict[str, str | bool | int | float]
    for i in range(study_config.optimization.num_retries_unique_params):
        params = {
            "prompt_name": trial.suggest_categorical(
                "prompt_name",
                search_space.prompt_names,
            ),
            "llm": trial.suggest_categorical(
                "llm",
                search_space.llms,
            ),
            "rag_mode": trial.suggest_categorical(
                "rag_mode",
                search_space.rag_modes,
            ),
            "embedding_models": trial.suggest_categorical(
                "embedding_models",
                search_space.embedding_models,
            ),
        }
        if params["rag_mode"] == "rag":
            params.update(search_space.splitter.sample(trial))

        flow_name = "LlamaIndexReactRAGAgentFlow"
        trial.set_user_attr("flow_name", flow_name)
        trial.set_user_attr(
            "datasets", ", ".join(ds.xname for ds in study_config.datasets)
        )

        if not study_config.optimization.skip_existing:
            logger.info("Using generated parameter combination without check")
            break
        if not trial_exists(trial.study.study_name, params):
            logger.info(
                "Found novel parameter combination after %i retries: %s",
                i,
                str(params),
            )
            break

    obj1, obj2, metrics, flow_json = evaluate(params, study_config)
    set_trial(trial, study_config, params, False, metrics, flow_json)  # type: ignore
    return obj1, obj2


@ray.remote
def seed_baselines(
    baseline: T.Dict[str, T.Any], study_config: AgentStudyConfig
) -> None:
    try:
        _seed_baselines(baseline, study_config)
    except Exception:
        logger.exception("Baseline had an unhandled exception. Skipping this baseline")


def _seed_baselines(
    baseline: T.Dict[str, T.Any], study_config: AgentStudyConfig
) -> None:
    from syftr.configuration import cfg
    from syftr.tuner.core import set_trial

    study_name = study_config.name
    search_space = study_config.search_space
    study = optuna.load_study(
        study_name=study_name, storage=cfg.database.get_optuna_storage()
    )
    if trial_exists(study_name, baseline, cfg.database.get_optuna_storage()):
        logger.warning(
            "Skipping evaluation of baseline which already exists in study '%s': %s",
            study_name,
            str(baseline),
        )
        return
    logger.info("Seeding study '%s' with baseline: %s", study_name, str(baseline))
    distributions: T.Dict[str, optuna.distributions.BaseDistribution] = {
        "llm": optuna.distributions.CategoricalDistribution(search_space.llms),
        "rag_mode": optuna.distributions.CategoricalDistribution(
            search_space.rag_modes
        ),
        "prompt_name": optuna.distributions.CategoricalDistribution(
            search_space.prompt_names
        ),
    }
    if baseline["rag_mode"] == "rag":
        distributions.update(**search_space.splitter.build_distributions())

    obj1, obj2, metrics, flow_json = evaluate(baseline, study_config)
    trial = optuna.create_trial(
        values=[obj1, obj2],
        params=baseline,
        distributions=distributions,
    )
    set_trial(trial, study_config, baseline, True, metrics, flow_json)  # type: ignore
    study.add_trial(trial)
    logger.debug("Seeding added trial with params: %s", trial.params)


def main():
    from syftr.configuration import cfg

    parser = argparse.ArgumentParser()
    default = get_default_study_name()
    parser.add_argument(
        "--study-config",
        help=f"Path to study config yaml (default: {default})",
        default=default,
    )
    parser.add_argument(
        "--remote",
        help="Use remote Ray cluster",
        default=False,
        action="store_true",
    )
    args = parser.parse_args()

    study_config_file = Path(args.study_config)
    study_config = AgentStudyConfig.from_file(study_config_file)
    cfg.ray.local = False if args.remote else cfg.ray.local

    ray_init()

    set_baselines(study_config)
    logger.info("Running study: %s", study_config.name)
    optimization = StudyRunner(
        objective=objective,
        study_config=study_config,
        seeder=seed_baselines,
    )
    optimization.run()


if __name__ == "__main__":
    main()



import hashlib
import json
from contextlib import contextmanager
from typing import Dict, Union

import pals
from flock import LOCK_EX, Flock

from syftr.configuration import cfg

LOCKER = pals.Locker("syftr-locks", create_engine_callable=cfg.database.get_engine)


@contextmanager
def distributed_lock(
    key_data: Union[Dict, str], timeout_s: int = 60 * 60 * 12, host_only: bool = False
):
    assert key_data, f"key_data must not be empty! Got {key_data=}"
    if isinstance(key_data, Dict):
        key_data = hashlib.sha1(
            json.dumps(key_data, sort_keys=True).encode("utf-8")
        ).hexdigest()

    if host_only:
        key_path = key_data.replace("/", "-")
        with open(cfg.paths.lock_dir / key_path, "w") as lockfile:
            with Flock(lockfile, LOCK_EX):
                yield key_data
    else:
        lock = LOCKER.lock(key_data)
        lock.acquire_timeout = timeout_s * 1000
        with lock:
            yield key_data



import typing as T
from functools import partial
from pathlib import Path

import datasets
import llama_index.core.instrumentation as instrument
import pandas as pd
import pytest
from fsspec.implementations.local import LocalFileSystem
from llama_index.core import Document, Settings, SimpleDirectoryReader
from llama_index.core.callbacks import CallbackManager, LlamaDebugHandler
from llama_index.core.evaluation import CorrectnessEvaluator
from llama_index.core.llms.llm import LLM
from llama_index.core.node_parser import SentenceSplitter
from llama_index.core.retrievers import BaseRetriever
from llama_index.core.storage.docstore.types import BaseDocumentStore
from llama_index.llms.azure_openai import AzureOpenAI

from syftr.configuration import AzureOpenAILLM, LLMCostTokens, cfg
from syftr.configuration import Settings as CfgSettings
from syftr.flows import Flow, RAGFlow, ReActAgentFlow
from syftr.huggingface_helper import get_embedding_model
from syftr.llm import get_llm, load_configured_llms
from syftr.retrievers.build import _build_dense_index, build_rag_retriever
from syftr.storage import HotPotQAHF, PartitionMap, QAPair, SyftrQADataset
from syftr.studies import (
    EmbeddingDeviceType,
    OptimizationConfig,
    ParamDict,
    StudyConfig,
    TimeoutConfig,
)
from syftr.templates import get_template

dispatcher = instrument.get_dispatcher()


@pytest.fixture(scope="session")
def bge_small():
    model, _ = get_embedding_model(
        "BAAI/bge-small-en-v1.5", use_hf_endpoint_models=False
    )
    return model


@pytest.fixture(scope="session")
def gpt_4o_mini() -> T.Tuple[AzureOpenAI, str]:
    llm = get_llm("gpt-4o-mini")
    assert isinstance(llm, AzureOpenAI), (
        "Expected AzureOpenAI instance for 'gpt-4o-mini'"
    )
    return llm, "gpt-4o-mini"


@pytest.fixture(scope="session")
def basic_flow(gpt_4o_mini):
    llm, _ = gpt_4o_mini
    return Flow(response_synthesizer_llm=llm, template=get_template("default"))


@pytest.fixture(
    scope="session",
    params=[
        (HotPotQAHF, "dev", "sample", "A collection of various trivia from Wikipedia"),
    ],
)
def real_dataset(request) -> SyftrQADataset:
    dataset_class, subset, partition, description = request.param
    if partition is not None:
        p_map = PartitionMap(test=partition)
    else:
        p_map = PartitionMap()
    return dataset_class(subset=subset, partition_map=p_map, description=description)


@pytest.fixture(scope="session")
def tiny_dataset() -> SyftrQADataset:
    class TinyDataset(SyftrQADataset):
        xname: str = "tiny-wiki-llm-page"  # type: ignore

        path_root: Path = cfg.paths.root_dir / "tests/functional/data/datasets"  # type: ignore
        dataset_dir: str = "partitioned"
        grounding_data_path: str = "grounding_data"
        examples_data_path: str = "shouldnotexist"

        def _row_to_qapair(self, row: T.Dict[str, T.Any]) -> QAPair:
            """Dataset-specific conversion of row to QAPair struct.

            Invoked by iter_examples.

            Default implementation assumes row is already in QAPair format.
            """
            return QAPair(**row)

        def iter_examples(self, partition="test", use_ray=False) -> T.Iterator[QAPair]:
            examples = self.load_examples()
            _iter: T.Callable[[], T.Iterator[T.Dict[str, T.Any]]]
            if use_ray:
                # Ray ray.datasets.Dataset.iter_rows
                _iter = examples.iter_rows
            else:
                # Huggingface datasets.Dataset.iter
                assert isinstance(examples, datasets.Dataset)

                def _iter():
                    for row in examples.iter(batch_size=1):
                        yield {key: value[0] for key, value in row.items()}

            for row in _iter():
                yield self._row_to_qapair(row)

        def iter_grounding_data(self, partition="notused") -> T.Iterator[Document]:
            """Iterate over grounding data examples."""
            grounding_data = self.load_grounding_data()
            # _iter: T.Callable[[], T.Iterator[Document]]

            def _iter():
                return iter(grounding_data)

            for row in _iter():
                yield row

        @property
        def storage(self) -> LocalFileSystem:
            return LocalFileSystem()

        def load_examples(
            self,
        ) -> datasets.Dataset:
            data = [
                {
                    "question": "What are some commonly used benchmarking datasets?",
                    "answer": "TruthfulQA, Web Questions, TriviaQA, and SQuAD",
                    "id": "1",
                    "context": {},
                    "supporting_facts": [
                        "Some examples of commonly used question answering datasets include TruthfulQA, Web Questions, TriviaQA, and SQuAD."
                    ],
                    "difficulty": "easy",
                    "qtype": "simple",
                },
                {
                    "question": "How much compute do you need to train llama 3.1",
                    "answer": "The 405B version of Llama 3.1 took 31 million hours on H100-80GB GPUs.",
                    "id": "2",
                    "context": {},
                    "supporting_facts": [
                        "405B version took 31 million hours on H100-80GB, at 3.8E25 FLOPs."
                    ],
                    "difficulty": "easy",
                    "qtype": "simple",
                },
            ]
            return datasets.Dataset.from_pandas(pd.DataFrame(data))

        def load_grounding_data(
            self,
            **load_kwargs,
        ) -> T.List[Document]:
            reader = SimpleDirectoryReader(
                input_dir="tests/functional/data/datasets/partitioned/tiny-wiki-llm-page/grounding_data/test",
            )
            return reader.load_data(**load_kwargs)

    return TinyDataset()


@pytest.fixture(scope="session")
def real_dataset_study_config(request, real_dataset) -> StudyConfig:
    device: EmbeddingDeviceType = (
        "cuda" if request.config.getoption("--gpu") else "onnx-cpu"
    )
    study_config = StudyConfig(
        name=f"functional_test_{real_dataset.name}",
        dataset=real_dataset,
        toy_mode=True,
        optimization=OptimizationConfig(embedding_device=device),
    )
    study_config.optimization.use_hf_embedding_models = False
    study_config.optimization.num_eval_samples = 10
    return study_config


@pytest.fixture(scope="session")
def tiny_dataset_study_config(hotpot_toy_study_config, tiny_dataset) -> StudyConfig:
    config: StudyConfig = hotpot_toy_study_config.copy()
    config.dataset = tiny_dataset
    config.optimization.use_hf_embedding_models = False
    return config


@pytest.fixture(scope="session")
def real_sparse_retriever(
    real_dataset_study_config,
) -> T.Tuple[BaseRetriever, BaseDocumentStore, StudyConfig]:
    params: ParamDict = {
        "rag_llm_name": "gpt-4o-mini",
        "rag_mode": "rag",
        "response_synthesizer_llm_name": "gpt-4o-mini",
        "rag_method": "sparse",
        "rag_top_k": 5,
        "rag_query_decomposition_enabled": True,
        "rag_query_decomposition_num_queries": 2,
        "rag_query_decomposition_llm_name": "gemini-flash2",
        "rag_fusion_mode": "simple",
        "splitter_method": "sentence",
        "splitter_chunk_exp": 9,
        "splitter_chunk_overlap_frac": 0.5,
    }
    retriever, docstore = build_rag_retriever(real_dataset_study_config, params)
    return retriever, docstore, real_dataset_study_config


@pytest.fixture(scope="session")
def real_dense_retriever(
    real_dataset_study_config,
) -> T.Tuple[BaseRetriever, BaseDocumentStore, StudyConfig]:
    params: ParamDict = {
        "rag_llm_name": "gpt-4o-mini",
        "rag_mode": "rag",
        "response_synthesizer_llm_name": "gpt-4o-mini",
        "rag_method": "dense",
        "rag_embedding_model": "BAAI/bge-small-en-v1.5",
        "rag_query_decomposition_enabled": False,
        "rag_top_k": 5,
        "splitter_method": "recursive",
        "splitter_chunk_exp": 9,
        "splitter_chunk_overlap_frac": 0.5,
    }
    retriever, docstore = build_rag_retriever(real_dataset_study_config, params)
    return retriever, docstore, real_dataset_study_config


@pytest.fixture(scope="session")
def tiny_sparse_retriever(
    tiny_dataset_study_config,
) -> T.Tuple[BaseRetriever, BaseDocumentStore]:
    params: ParamDict = {
        "rag_llm_name": "gpt-4o-mini",
        "rag_mode": "rag",
        "response_synthesizer_llm_name": "gpt-4o-mini",
        "rag_method": "sparse",
        "rag_top_k": 5,
        "rag_query_decomposition_enabled": False,
        "splitter_method": "sentence",
        "splitter_chunk_exp": 9,
        "splitter_chunk_overlap_frac": 0.5,
    }
    return build_rag_retriever(tiny_dataset_study_config, params)


@pytest.fixture(scope="session")
def tiny_dense_retriever(
    tiny_dataset_study_config,
) -> T.Tuple[BaseRetriever, BaseDocumentStore]:
    params: ParamDict = {
        "rag_llm_name": "gpt-4o-mini",
        "rag_mode": "rag",
        "response_synthesizer_llm_name": "gpt-4o-mini",
        "rag_method": "dense",
        "rag_embedding_model": "BAAI/bge-small-en-v1.5",
        "rag_query_decomposition_enabled": False,
        "rag_top_k": 5,
        "splitter_method": "recursive",
        "splitter_chunk_exp": 9,
        "splitter_chunk_overlap_frac": 0.5,
    }
    return build_rag_retriever(tiny_dataset_study_config, params)


@pytest.fixture(scope="session")
def real_hybrid_retriever(
    real_dataset_study_config,
) -> T.Tuple[BaseRetriever, BaseDocumentStore, StudyConfig]:
    params: ParamDict = {
        "rag_llm_name": "gpt-4o-mini",
        "rag_mode": "rag",
        "rag_method": "hybrid",
        "response_synthesizer_llm_name": "gpt-4o-mini",
        "rag_embedding_model": "BAAI/bge-small-en-v1.5",
        "rag_top_k": 5,
        "rag_query_decomposition_enabled": True,
        "rag_query_decomposition_llm_name": "gpt-4o-mini",
        "rag_query_decomposition_num_queries": 2,
        "rag_fusion_mode": "simple",
        "rag_hybrid_bm25_weight": 0.5,
        "splitter_method": "sentence",
        "splitter_chunk_exp": 9,
        "splitter_chunk_overlap_frac": 0.5,
    }
    retriever, docstore = build_rag_retriever(real_dataset_study_config, params)
    return retriever, docstore, real_dataset_study_config


@pytest.fixture(scope="session")
def rag_template():
    return get_template("default", with_context=True)


@pytest.fixture(scope="session")
def few_shot_rag_template():
    return get_template("default", with_context=True, with_few_shot_prompt=True)


@pytest.fixture(scope="session")
def bge_small_no_hf():
    model, _ = get_embedding_model(
        "BAAI/bge-small-en-v1.5", use_hf_endpoint_models=False
    )
    return model


@pytest.fixture(scope="session")
def correctness_evaluator(gpt_4o_mini):
    llm, _ = gpt_4o_mini
    evaluator = CorrectnessEvaluator(llm)
    evaluator.evaluate = dispatcher.span(evaluator.evaluate)
    evaluator.aevaluate = dispatcher.span(evaluator.aevaluate)
    return evaluator


@pytest.fixture
def hotpot_dense_flow(hotpot_dense_retriever, gpt_4o_mini, rag_template) -> RAGFlow:
    llm, _ = gpt_4o_mini
    hotpot_dense_retriever, _ = hotpot_dense_retriever
    return RAGFlow(
        retriever=hotpot_dense_retriever,
        response_synthesizer_llm=llm,
        template=rag_template,
    )


@pytest.fixture
def hotpot_fusion_flow(hotpot_fusion_retriever, gpt_4o_mini, rag_template) -> RAGFlow:
    llm, _ = gpt_4o_mini
    hotpot_fusion_retriever, _ = hotpot_fusion_retriever
    return RAGFlow(
        retriever=hotpot_fusion_retriever,
        response_synthesizer_llm=llm,
        template=rag_template,
    )


@pytest.fixture
def hotpot_reranker_flow(hotpot_dense_retriever, gpt_4o_mini, rag_template) -> RAGFlow:
    llm, _ = gpt_4o_mini
    retriever, _ = hotpot_dense_retriever
    return RAGFlow(
        retriever=retriever,
        response_synthesizer_llm=llm,
        template=rag_template,
        reranker_llm=llm,
        reranker_top_k=5,
    )


@pytest.fixture(scope="session")
def hotpot_toy_study_config() -> StudyConfig:
    return StudyConfig.from_file(cfg.paths.test_studies_dir / "test-hotpot-toy.yaml")


@pytest.fixture(scope="session")
def hotpot_sparse_retriever(
    hotpot_toy_study_config,
) -> T.Tuple[BaseRetriever, BaseDocumentStore]:
    params: ParamDict = {
        "rag_mode": "rag",
        "rag_method": "sparse",
        "rag_top_k": 5,
        "rag_query_decomposition_enabled": False,
        "splitter_method": "sentence",
        "splitter_chunk_exp": 9,
        "splitter_chunk_overlap_frac": 0.5,
    }
    return build_rag_retriever(hotpot_toy_study_config, params)


@pytest.fixture(scope="session")
def hotpot_dense_retriever(
    hotpot_toy_study_config,
) -> T.Tuple[BaseRetriever, BaseDocumentStore]:
    params: ParamDict = {
        "response_synthesizer_llm_name": "gpt-4o-mini",
        "rag_mode": "rag",
        "rag_method": "dense",
        "rag_embedding_model": "BAAI/bge-small-en-v1.5",
        "rag_top_k": 5,
        "rag_query_decomposition_enabled": False,
        "splitter_method": "sentence",
        "splitter_chunk_exp": 9,
        "splitter_chunk_overlap_frac": 0.5,
    }
    return build_rag_retriever(hotpot_toy_study_config, params)


@pytest.fixture(scope="session")
def hotpot_dense_retriever_recursive(
    hotpot_toy_study_config,
) -> T.Tuple[BaseRetriever, BaseDocumentStore]:
    params: ParamDict = {
        "rag_mode": "rag",
        "rag_method": "dense",
        "rag_embedding_model": "BAAI/bge-small-en-v1.5",
        "rag_top_k": 5,
        "rag_query_decomposition_enabled": False,
        "splitter_method": "recursive",
        "splitter_chunk_exp": 9,
        "splitter_chunk_overlap_frac": 0.5,
    }
    return build_rag_retriever(hotpot_toy_study_config, params)


@pytest.fixture(scope="session")
def hotpot_fusion_retriever(
    hotpot_toy_study_config,
) -> T.Tuple[BaseRetriever, BaseDocumentStore]:
    params: ParamDict = {
        "rag_mode": "rag",
        "rag_method": "fusion",
        "rag_embedding_model": "BAAI/bge-small-en-v1.5",
        "rag_top_k": 5,
        "rag_query_decomposition_enabled": False,
        "rag_fusion_llm_name": "gpt-4o-mini",
        "rag_fusion_num_queries": 2,
        "rag_fusion_mode": "simple",
        "splitter_method": "sentence",
        "splitter_chunk_exp": 9,
        "splitter_chunk_overlap_frac": 0.5,
    }
    return build_rag_retriever(hotpot_toy_study_config, params)


@pytest.fixture(scope="session")
def llama_debug():
    llama_debug = LlamaDebugHandler(print_trace_on_end=True)
    callback_manager = CallbackManager([llama_debug])
    Settings.callback_manager = callback_manager
    return llama_debug


@pytest.fixture
def tiny_flow(tiny_dense_retriever, gpt_4o_mini, rag_template) -> RAGFlow:
    llm, _ = gpt_4o_mini
    retriever, _ = tiny_dense_retriever
    return RAGFlow(
        retriever=retriever,
        response_synthesizer_llm=llm,
        template=rag_template,
        params={},
    )


@pytest.fixture
def hotpot_dev(hotpot_toy_study_config):
    return hotpot_toy_study_config.dataset


@pytest.fixture
def build_index_that_will_time_out(bge_small_no_hf, hotpot_dev):
    """Only use if you want to test the embeddings timeout functionality."""
    splitter = SentenceSplitter(chunk_size=512, chunk_overlap=128)
    transforms = [splitter]

    bge_small_no_hf.reset_timeouts(
        timeout_config=TimeoutConfig(
            embedding_timeout_active=True,
            embedding_max_time=2,
            embedding_min_chunks_to_process=20,
            embedding_min_time_to_process=1,
        )
    )

    return partial(
        _build_dense_index,
        list(hotpot_dev.iter_grounding_data()),
        transforms,
        bge_small_no_hf,
    )


@pytest.fixture
def build_index_that_will_skip_time_out(bge_small_no_hf, hotpot_dev):
    """Only use if you want to test the embeddings timeout functionality."""
    splitter = SentenceSplitter(chunk_size=512, chunk_overlap=128)
    transforms = [splitter]

    bge_small_no_hf.reset_timeouts(
        timeout_config=TimeoutConfig(
            embedding_timeout_active=False,
            embedding_max_time=2,
            embedding_min_chunks_to_process=20,
            embedding_min_time_to_process=1,
        )
    )

    return partial(
        _build_dense_index,
        list(hotpot_dev.iter_grounding_data()),
        transforms,
        bge_small_no_hf,
    )


@pytest.fixture
def build_index_that_will_not_time_out(bge_small_no_hf, hotpot_dev):
    """Only use if you want to test the embeddings timeout functionality."""
    splitter = SentenceSplitter(chunk_size=512, chunk_overlap=128)
    transforms = [splitter]

    bge_small_no_hf.reset_timeouts(
        timeout_config=TimeoutConfig(
            embedding_max_time=1000,
            embedding_min_chunks_to_process=50,
            embedding_min_time_to_process=2,
        )
    )

    return partial(
        _build_dense_index,
        list(hotpot_dev.iter_grounding_data()),
        transforms,
        bge_small_no_hf,
    )


@pytest.fixture(scope="session")
def sentence_splitter() -> SentenceSplitter:
    return SentenceSplitter(
        chunk_size=256,
        chunk_overlap=64,
    )


@pytest.fixture
def hotpot_study_config():
    study_config = StudyConfig.from_file(
        "tests/functional/data/studies/test-hotpot-toy.yaml"
    )
    return study_config


@pytest.fixture(scope="session")
def react_agent_flow(
    hotpot_dense_retriever, hotpot_toy_study_config, gpt_4o_mini, rag_template
) -> T.Tuple[ReActAgentFlow, StudyConfig]:
    llm, _ = gpt_4o_mini
    hotpot_dense_retriever, _ = hotpot_dense_retriever
    return (
        ReActAgentFlow(
            retriever=hotpot_dense_retriever,
            response_synthesizer_llm=llm,
            subquestion_engine_llm=llm,
            subquestion_response_synthesizer_llm=llm,
            template=rag_template,
            dataset_name=hotpot_toy_study_config.dataset.name,
            dataset_description=hotpot_toy_study_config.dataset.description,
        ),
        hotpot_toy_study_config,
    )


def config_with_models() -> CfgSettings:
    custom_config = CfgSettings(
        generative_models={
            "test_gpt_4o_mini": AzureOpenAILLM(
                deployment_name="gpt-4o-mini",
                model_name="gpt-4o-mini",
                api_version="2024-06-01",
                additional_kwargs={"user": "syftr"},
                cost=LLMCostTokens(input=0.15, output=0.60),
            ),
            "test_gpt_4o": AzureOpenAILLM(
                deployment_name="gpt-4o",
                model_name="gpt-4o",
                api_version="2024-06-01",
                additional_kwargs={"user": "syftr"},
                cost=LLMCostTokens(input=2.5, output=10.00),
            ),
        }
    )
    custom_config.generative_models = {
        key: value
        for key, value in custom_config.generative_models.items()
        if "test_" in key
    }
    return custom_config


@pytest.fixture(
    params=[name for name in config_with_models().generative_models.keys()],
    ids=[name for name in config_with_models().generative_models.keys()],
)
def configured_models(request) -> T.Tuple[str, LLM]:
    name = request.param
    llms = load_configured_llms(config_with_models())
    return name, llms[name]



import logging
from logging.handlers import RotatingFileHandler

import colorlog

from syftr.configuration import cfg

if cfg.logging.use_colors:
    logger = colorlog.getLogger(cfg.logging.name)
else:
    logger = logging.getLogger(cfg.logging.name)


def _create_default_handler(handler_name: str) -> logging.Handler:
    """Gives the project default logging handler"""
    if cfg.logging.use_colors:
        new_handler = colorlog.StreamHandler()
        new_handler.setFormatter(colorlog.ColoredFormatter(cfg.logging.color_format))
    else:
        new_handler = logging.StreamHandler()
        new_handler.setFormatter(logging.Formatter(cfg.logging.normal_format))
    new_handler.set_name(handler_name)
    return new_handler


def has_handler(logger_instance: logging.Logger, handler_name: str) -> bool:
    if not logger.handlers:
        return False
    for han in logger_instance.handlers:
        if han.name == handler_name:
            return True
    return False


def _create_file_handler(handler_name: str, file_name: str) -> logging.Handler:
    new_handler = RotatingFileHandler(
        file_name, maxBytes=2 * 1024 * 1024, backupCount=5
    )
    normal_formatter = logging.Formatter(cfg.logging.normal_format)
    new_handler.setFormatter(normal_formatter)
    new_handler.set_name(handler_name)
    return new_handler


def add_file_handler(
    logger_instance: logging.Logger,
    file_name: str = cfg.logging.filename,
    handler_name: str = "file_handler",
):
    file_handler = None
    if logger.handlers:
        for han in logger_instance.handlers:
            if han.name == handler_name:
                file_handler = han
    if file_handler is None:
        file_handler = _create_file_handler(handler_name, file_name)
    logger.addHandler(file_handler)


def add_default_handler(logger_instance: logging.Logger):
    if not has_handler(logger_instance, "default_handler"):
        default_handler = _create_default_handler("default_handler")
        logger.addHandler(default_handler)


add_default_handler(logger)
# add_file_handler(logger)
logger.setLevel(cfg.logging.level)


def io_logger(func):
    """Decorator that returns a wrapped function for logging"""

    def wrapper(*args, **kwargs):
        """Wrapped function for logging"""
        logger.debug("%s input:", func.__name__)
        for value in args:
            logger.debug("\t%s", value)
        for parameter, value in kwargs.items():
            logger.debug("\t%s=%s", parameter, value)
        result = func(*args, **kwargs)
        if result:
            logger.debug("%s output:", func.__name__)
            logger.debug("\t%s", result)
        return result

    return wrapper


if __name__ == "__main__":
    logger.info("Logging module loaded")



import optuna

from syftr.configuration import cfg

# STUDY_NAME = "run9"
STUDY_NAME = "hotpot-test-fullwiki-toy"

storage = cfg.database.get_optuna_storage()
optuna.delete_study(study_name=STUDY_NAME, storage=storage)



from aiolimiter import AsyncLimiter

from syftr.flows import RAGFlow
from syftr.prompt_optimization import optimize_prompt
from syftr.storage import SyftrQADataset


def test_basic_optimization(tiny_flow: RAGFlow, tiny_dataset: SyftrQADataset):
    raw_data = list(tiny_dataset.iter_examples())
    train = raw_data
    test = raw_data
    rate_limiter = AsyncLimiter(10, 10)
    resulting_prompt = optimize_prompt(
        tiny_flow,
        "gpt-4o-mini",
        "gpt-4o-mini",
        train,
        test,
        rate_limiter,
        num_epochs=3,
    )
    assert resulting_prompt



import random
import time
import typing as T
from collections import Counter
from dataclasses import dataclass, field
from enum import Enum
from uuid import uuid4

import llama_index.core.instrumentation as instrument
from llama_index.agent.introspective import (
    IntrospectiveAgentWorker,
    ToolInteractiveReflectionAgentWorker,
)
from llama_index.agent.introspective.reflective.tool_interactive_reflection import (
    StoppingCallable,
)
from llama_index.agent.lats import LATSAgentWorker
from llama_index.core import (
    PromptTemplate,
    QueryBundle,
    Response,
    get_response_synthesizer,
)
from llama_index.core.agent import (
    AgentChatResponse,
    AgentRunner,
    FunctionCallingAgentWorker,
    ReActAgent,
)
from llama_index.core.agent.react.formatter import ReActChatFormatter
from llama_index.core.evaluation import EvaluationResult
from llama_index.core.indices.query.query_transform.base import HyDEQueryTransform
from llama_index.core.llms import ChatMessage, CompletionResponse, MessageRole
from llama_index.core.llms.function_calling import FunctionCallingLLM
from llama_index.core.llms.llm import LLM
from llama_index.core.postprocessor import LLMRerank, PrevNextNodePostprocessor
from llama_index.core.postprocessor.types import BaseNodePostprocessor
from llama_index.core.prompts import PromptType
from llama_index.core.query_engine import (
    BaseQueryEngine,
    RetrieverQueryEngine,
    SubQuestionQueryEngine,
    TransformQueryEngine,
)
from llama_index.core.response_synthesizers.type import ResponseMode
from llama_index.core.retrievers import BaseRetriever
from llama_index.core.schema import NodeWithScore
from llama_index.core.storage.docstore.types import BaseDocumentStore
from llama_index.core.tools import BaseTool, FunctionTool, QueryEngineTool, ToolMetadata
from llama_index.packs.agents_coa import CoAAgentPack
from numpy import ceil

from syftr.configuration import cfg
from syftr.instrumentation.arize import instrument_arize
from syftr.instrumentation.tokens import LLMCallData, TokenTrackingEventHandler
from syftr.llm import get_nothink_str, get_tokenizer
from syftr.logger import logger
from syftr.output_parsers.judge import parse_correctness_evaluation
from syftr.prompts.judge import (
    DEFAULT_JUDGE_QUERY_PROMPT_TEMPLATE,
    DEFAULT_JUDGE_SYSTEM_PROMPT,
)
from syftr.studies import get_critique_template, get_react_template

dispatcher = instrument.get_dispatcher()
_event_handler = TokenTrackingEventHandler()
dispatcher.add_event_handler(_event_handler)
dispatcher.add_span_handler(_event_handler._span_handler)

if cfg.instrumentation.tracing_enabled:
    instrument_arize()


@dataclass(kw_only=True)
class Flow:
    response_synthesizer_llm: LLM | FunctionCallingLLM | None = None
    template: str | None = None
    get_examples: T.Callable | None = None
    name: str = "Generator Flow"
    params: dict | None = None
    enforce_full_evaluation: bool = False
    use_reasoning: bool | None = None

    _llm_call_data: T.Dict[str, T.List[LLMCallData]] = field(default_factory=dict)

    def __repr__(self):
        return f"{self.name}: {self.params}"

    @property
    def verbose(self):
        log_level = logger.level
        if log_level <= 20:
            return True
        return False

    @property
    def prompt_template(self) -> PromptTemplate:
        if self.template is None:
            raise ValueError("Flow template not set. Cannot create prompt template.")
        prompt_template = None
        function_mappings = None
        if self.get_examples is not None:
            function_mappings = {"few_shot_examples": self.get_examples}

        logger.debug("Creating prompt template from '%s'", self.template)
        prompt_template = PromptTemplate(
            template=self.template,
            prompt_type=PromptType.QUESTION_ANSWER,
            function_mappings=function_mappings,
        )

        return prompt_template

    def get_prompt(self, query) -> str:
        if self.template is None:
            return query

        if self.get_examples is None:
            return self.template.format(query_str=query)

        examples = self.get_examples(query)
        assert examples, "No examples found for few-shot prompting"

        return self.template.format(
            query_str=query,
            few_shot_examples=examples,
        )

    def set_thinking(self, query: str) -> str:
        if self.use_reasoning is None:
            return query
        no_think = get_nothink_str(self.response_synthesizer_llm)
        thinking = "/think" if self.use_reasoning else no_think
        return f"{query} {thinking}"

    def generate(
        self, query: str
    ) -> T.Tuple[CompletionResponse, float, T.List[LLMCallData]]:
        invocation_id = uuid4().hex
        self._llm_call_data[invocation_id] = []
        response, duration = self._generate(query, invocation_id)
        call_data = self._llm_call_data.pop(invocation_id)
        return response, duration, call_data

    @dispatcher.span
    def _generate(
        self, query: str, invocation_id: str
    ) -> T.Tuple[CompletionResponse, float]:
        assert self.response_synthesizer_llm is not None, (
            "Response synthesizer LLM is not set. Cannot generate."
        )
        start_time = time.perf_counter()
        query = self.set_thinking(query)
        prompt = self.get_prompt(query)
        response: CompletionResponse = self.response_synthesizer_llm.complete(prompt)
        duration = time.perf_counter() - start_time
        return response, duration

    async def agenerate(
        self, query: str
    ) -> T.Tuple[CompletionResponse, float, T.List[LLMCallData]]:
        invocation_id = uuid4().hex
        self._llm_call_data[invocation_id] = []
        response, duration = await self._agenerate(query, invocation_id)
        call_data = self._llm_call_data.pop(invocation_id)
        return response, duration, call_data

    @dispatcher.span
    async def _agenerate(
        self, query: str, invocation_id: str
    ) -> T.Tuple[CompletionResponse, float]:
        assert self.response_synthesizer_llm is not None, (
            "Response synthesizer LLM is not set. Cannot generate."
        )
        start_time = time.perf_counter()
        query = self.set_thinking(query)
        prompt = self.get_prompt(query)
        response: CompletionResponse = await self.response_synthesizer_llm.acomplete(
            prompt
        )
        duration = time.perf_counter() - start_time
        return response, duration


@dataclass(kw_only=True)
class RetrieverFlow(Flow):
    """Flow that only retrieves documents."""

    retriever: BaseRetriever
    docstore: BaseDocumentStore | None = None
    hyde_llm: LLM | None = None
    additional_context_num_nodes: int = 0
    name: str = "Retriever Only Flow"

    def __repr__(self):
        return f"{self.name}: {self.params}"

    @property
    def query_engine(self) -> BaseQueryEngine:
        node_postprocessors: T.List[BaseNodePostprocessor] = []
        if self.additional_context_num_nodes > 0:
            assert self.docstore is not None
            node_postprocessors.append(
                PrevNextNodePostprocessor(
                    docstore=self.docstore,
                    num_nodes=int(ceil(self.additional_context_num_nodes / 2)),
                    mode="both",
                )
            )
        response_synthesizer = get_response_synthesizer(
            llm=self.response_synthesizer_llm,
            response_mode=ResponseMode.COMPACT,
        )
        base_engine = RetrieverQueryEngine(
            retriever=self.retriever,
            response_synthesizer=response_synthesizer,
            node_postprocessors=node_postprocessors,
        )
        if self.hyde_llm is not None:
            hyde = HyDEQueryTransform(llm=self.hyde_llm, include_original=True)
            return TransformQueryEngine(base_engine, query_transform=hyde)
        return base_engine

    @property
    def tokenizer(self) -> T.Callable:
        return get_tokenizer(self.response_synthesizer_llm.model)  # type: ignore

    def generate(self, query: str, *args, **kwargs):
        raise NotImplementedError("RetrieverFlow does not support generation.")

    async def agenerate(self, query: str, *args, **kwargs):
        raise NotImplementedError("RetrieverFlow does not support generation.")

    @dispatcher.span
    def retrieve(self, query: str) -> T.Tuple[T.List[NodeWithScore], float]:
        start_time = time.perf_counter()
        query = self.set_thinking(query)
        qb = QueryBundle(query)
        if isinstance(self.query_engine, TransformQueryEngine):
            response = self.query_engine.query(qb)
            assert isinstance(response, Response), (
                f"Expected Response, got {type(response)=}"
            )
            retrieval_result = response.source_nodes
        else:
            retrieval_result = self.query_engine.retrieve(qb)
        duration = time.perf_counter() - start_time
        return retrieval_result, duration

    @dispatcher.span
    async def aretrieve(self, query: str) -> T.Tuple[T.List[NodeWithScore], float]:
        start_time = time.perf_counter()
        query = self.set_thinking(query)
        qb = QueryBundle(query)
        if isinstance(self.query_engine, TransformQueryEngine):
            response = await self.query_engine.aquery(qb)
            assert isinstance(response, Response), (
                f"Expected Response, got {type(response)=}"
            )
            retrieval_result = response.source_nodes
        else:
            assert hasattr(self.query_engine, "aretrieve"), (
                f"{self.query_engine} does not have 'aretrieve' method"
            )
            retrieval_result = await self.query_engine.aretrieve(qb)  # type: ignore
        duration = time.perf_counter() - start_time
        return retrieval_result, duration


@dataclass(kw_only=True)
class JudgeFlow(Flow):
    """Flow that evaluates whether a response is correct."""

    name: str = "Judge Flow"
    system_prompt: str = DEFAULT_JUDGE_SYSTEM_PROMPT
    query_prompt_template: str = DEFAULT_JUDGE_QUERY_PROMPT_TEMPLATE
    output_parser: T.Callable[[str, str], EvaluationResult] = (
        parse_correctness_evaluation
    )
    temperature: float = 0.0

    def __repr__(self):
        return f"{self.name}: {self.params}"

    @property
    def tokenizer(self) -> T.Callable:
        return get_tokenizer(self.response_synthesizer_llm.model)  # type: ignore

    def get_prompt(self, question, answer, response) -> str:  # type: ignore
        """Returns final prompt string to send in completion request."""
        query = self.query_prompt_template.format(
            question=question, answer=answer, response=response
        )
        return self.system_prompt + "\n" + query

    def generate(self, query: str, *args, **kwargs):
        raise NotImplementedError("JudgeFlow does not support generation.")

    async def agenerate(self, query: str, *args, **kwargs):
        raise NotImplementedError("JudgeFlow does not support generation.")

    def judge(
        self, question: str, answer: str, response: str
    ) -> T.Tuple[EvaluationResult, float, T.List[LLMCallData]]:
        invocation_id = uuid4().hex
        self._llm_call_data[invocation_id] = []
        start_time = time.perf_counter()
        response = self._judge(question, answer, response, invocation_id)
        duration = time.perf_counter() - start_time
        call_data = self._llm_call_data.pop(invocation_id)
        return response, duration, call_data

    @dispatcher.span
    def _judge(
        self,
        question: str,
        answer: str,
        response: str,
        invocation_id: str,
        response_synthesizer_llm: T.Optional[LLM | FunctionCallingLLM] = None,
    ) -> EvaluationResult:
        prompt = self.get_prompt(question, answer, response)
        llm = response_synthesizer_llm or self.response_synthesizer_llm
        assert llm is not None, "Response synthesizer LLM is not set. Cannot judge."
        llm.temperature = self.temperature  # type: ignore
        judge_response: CompletionResponse = llm.complete(prompt)
        result = self.output_parser(prompt, judge_response.text)
        return result

    async def ajudge(
        self, question: str, answer: str, response: str
    ) -> T.Tuple[EvaluationResult, float, T.List[LLMCallData]]:
        invocation_id = uuid4().hex
        self._llm_call_data[invocation_id] = []
        start_time = time.perf_counter()
        response = await self._ajudge(question, answer, response, invocation_id)
        duration = time.perf_counter() - start_time
        call_data = self._llm_call_data.pop(invocation_id)
        return response, duration, call_data

    @dispatcher.span
    async def _ajudge(
        self,
        question: str,
        answer: str,
        response: str,
        invocation_id: str,
        response_synthesizer_llm: T.Optional[LLM | FunctionCallingLLM] = None,
    ) -> EvaluationResult:
        prompt = self.get_prompt(question, answer, response)
        llm = response_synthesizer_llm or self.response_synthesizer_llm
        assert llm is not None, "Response synthesizer LLM is not set. Cannot judge."
        llm.temperature = self.temperature  # type: ignore
        judge_response: CompletionResponse = await llm.acomplete(prompt)
        result = self.output_parser(prompt, judge_response.text)
        return result


@dataclass(kw_only=True)
class MasterRMFlow(JudgeFlow):
    """Flow specific to Master-RM model."""

    name: str = "Judge Flow"
    system_prompt: str = (
        "Return YES if the Generated Answer is correct relative to the Reference Answer"
    )
    output_parser: T.Callable[[str, str], EvaluationResult] = (
        lambda q, r: EvaluationResult(
            query=q, response=r, passing=r.strip().lower() == "yes"
        )
    )
    temperature: float = 0.0


@dataclass(kw_only=True)
class ConsensusJudgeFlow(JudgeFlow):
    """Flow that evaluates whether a response is correct."""

    name: str = "Consensus Judge Flow"
    response_synthesizer_llms: T.List[LLM | FunctionCallingLLM]

    def judge(
        self, question: str, answer: str, response: str
    ) -> T.Tuple[EvaluationResult, float, T.List[LLMCallData]]:
        invocation_id = uuid4().hex
        self._llm_call_data[invocation_id] = []
        start_time = time.perf_counter()

        responses: T.List[EvaluationResult] = []
        for response_synthesizer_llm in self.response_synthesizer_llms:
            judge_response: EvaluationResult = self._judge(
                question, answer, response, invocation_id, response_synthesizer_llm
            )
            responses.append(judge_response)

        judge_response.passing = Counter([r.passing for r in responses]).most_common(1)[
            0
        ][0]
        float_scores = [
            r.score
            for r in responses
            if isinstance(r.score, float) and r.score is not None
        ]
        judge_response.score = (
            sum(float_scores) / len(float_scores) if float_scores else None
        )

        duration = time.perf_counter() - start_time
        call_data = self._llm_call_data.pop(invocation_id)
        return judge_response, duration, call_data

    async def ajudge(
        self, question: str, answer: str, response: str
    ) -> T.Tuple[EvaluationResult, float, T.List[LLMCallData]]:
        invocation_id = uuid4().hex
        self._llm_call_data[invocation_id] = []
        start_time = time.perf_counter()

        responses: T.List[EvaluationResult] = []
        for response_synthesizer_llm in self.response_synthesizer_llms:
            judge_response: EvaluationResult = await self._ajudge(
                question, answer, response, invocation_id, response_synthesizer_llm
            )
            responses.append(judge_response)

        judge_response.passing = Counter([r.passing for r in responses]).most_common(1)[
            0
        ][0]
        float_scores = [
            r.score
            for r in responses
            if isinstance(r.score, float) and r.score is not None
        ]
        judge_response.score = (
            sum(float_scores) / len(float_scores) if float_scores else None
        )

        duration = time.perf_counter() - start_time
        call_data = self._llm_call_data.pop(invocation_id)
        return judge_response, duration, call_data


@dataclass(kw_only=True)
class RandomJudgeFlow(JudgeFlow):
    """Flow that evaluates whether a response is correct."""

    name: str = "Random Judge Flow"
    response_synthesizer_llms: T.List[LLM | FunctionCallingLLM]

    def judge(
        self, question: str, answer: str, response: str
    ) -> T.Tuple[EvaluationResult, float, T.List[LLMCallData]]:
        invocation_id = uuid4().hex
        self._llm_call_data[invocation_id] = []
        start_time = time.perf_counter()
        response_synthesizer_llm = random.choice(self.response_synthesizer_llms)
        response = self._judge(
            question, answer, response, invocation_id, response_synthesizer_llm
        )
        duration = time.perf_counter() - start_time
        call_data = self._llm_call_data.pop(invocation_id)
        return response, duration, call_data

    async def ajudge(
        self, question: str, answer: str, response: str
    ) -> T.Tuple[EvaluationResult, float, T.List[LLMCallData]]:
        invocation_id = uuid4().hex
        self._llm_call_data[invocation_id] = []
        start_time = time.perf_counter()
        response_synthesizer_llm = random.choice(self.response_synthesizer_llms)
        response = await self._ajudge(
            question, answer, response, invocation_id, response_synthesizer_llm
        )
        duration = time.perf_counter() - start_time
        call_data = self._llm_call_data.pop(invocation_id)
        return response, duration, call_data


@dataclass(kw_only=True)
class RAGFlow(Flow):
    retriever: BaseRetriever
    docstore: BaseDocumentStore | None = None
    hyde_llm: LLM | None = None
    reranker_llm: LLM | None = None
    reranker_top_k: int | None = None
    name: str = "RAG Flow"
    additional_context_num_nodes: int = 0

    def __repr__(self):
        return f"{self.name}: {self.params}"

    @property
    def query_engine(self) -> BaseQueryEngine:
        node_postprocessors: T.List[BaseNodePostprocessor] = []
        if self.reranker_llm is not None:
            assert self.reranker_top_k, (
                "Reranker enabled, need reranker_top_k param set"
            )
            node_postprocessors.append(
                LLMRerank(top_n=self.reranker_top_k, llm=self.reranker_llm)
            )
        if self.additional_context_num_nodes > 0:
            assert self.docstore is not None
            node_postprocessors.append(
                PrevNextNodePostprocessor(
                    docstore=self.docstore,
                    num_nodes=int(ceil(self.additional_context_num_nodes / 2)),
                    mode="both",
                )
            )
        response_synthesizer = get_response_synthesizer(
            llm=self.response_synthesizer_llm,
            response_mode=ResponseMode.COMPACT,
            text_qa_template=self.prompt_template,
        )
        retriever = RetrieverQueryEngine(
            retriever=self.retriever,
            response_synthesizer=response_synthesizer,
            node_postprocessors=node_postprocessors,
        )
        if self.hyde_llm is not None:
            hyde = HyDEQueryTransform(llm=self.hyde_llm, include_original=True)
            retriever = TransformQueryEngine(retriever, query_transform=hyde)  # type: ignore
        return retriever

    def retrieve(self, query: str) -> T.List[NodeWithScore]:
        return self.query_engine.retrieve(QueryBundle(query))

    async def aretrieve(self, query: str) -> T.List[NodeWithScore]:
        assert hasattr(self.query_engine, "aretrieve"), (
            f"{self.query_engine} does not have 'aretrieve' method"
        )
        return await self.query_engine.aretrieve(QueryBundle(query))  # type: ignore

    @dispatcher.span
    def _generate(
        self, query: str, invocation_id: str
    ) -> T.Tuple[CompletionResponse, float]:
        start_time = time.perf_counter()
        query = self.set_thinking(query)
        response = self.query_engine.query(query)
        assert isinstance(response, Response), (
            f"Expected Response, got {type(response)=}"
        )
        completion_response = CompletionResponse(
            text=str(response.response),
            additional_kwargs={
                "source_nodes": response.source_nodes,
                **(response.metadata or {}),  # type: ignore
            },
        )
        duration = time.perf_counter() - start_time
        return completion_response, duration

    @dispatcher.span
    async def _agenerate(
        self, query: str, invocation_id: str
    ) -> T.Tuple[CompletionResponse, float]:
        start_time = time.perf_counter()
        query = self.set_thinking(query)
        response = await self.query_engine.aquery(query)
        assert isinstance(response, Response), (
            f"Expected Response, got {type(response)=}"
        )
        assert isinstance(response.response, str), (
            f"Expected str, got {type(response.response)=}"
        )
        completion_response = CompletionResponse(
            text=str(response.response),
            additional_kwargs={
                "source_nodes": response.source_nodes,
                **(response.metadata or {}),
            },
        )
        duration = time.perf_counter() - start_time
        return completion_response, duration


@dataclass(kw_only=True)
class SubQuestionRAGFlow(RAGFlow):
    dataset_name: str
    dataset_description: str
    subquestion_engine_llm: LLM
    subquestion_response_synthesizer_llm: LLM
    name: str = "Sub-Question Agentic RAG Flow"

    def __repr__(self):
        return f"{self.name}: {self.params}"

    @property
    def tools(self) -> T.List[QueryEngineTool]:
        # Build query engine like RAGFlow, since this class overrides query_engine
        node_postprocessors: T.List[BaseNodePostprocessor] = []
        if self.reranker_llm is not None:
            assert self.reranker_top_k, (
                "Reranker enabled, need reranker_top_k param set"
            )
            node_postprocessors.append(
                LLMRerank(top_n=self.reranker_top_k, llm=self.reranker_llm)
            )
        if self.additional_context_num_nodes > 0:
            assert self.docstore is not None
            node_postprocessors.append(
                PrevNextNodePostprocessor(
                    docstore=self.docstore,
                    num_nodes=int(ceil(self.additional_context_num_nodes / 2)),
                    mode="both",
                )
            )
        response_synthesizer = get_response_synthesizer(
            llm=self.subquestion_response_synthesizer_llm,
            response_mode=ResponseMode.COMPACT,
            text_qa_template=self.prompt_template,
        )
        retriever = RetrieverQueryEngine(
            retriever=self.retriever,
            response_synthesizer=response_synthesizer,
            node_postprocessors=node_postprocessors,
        )
        if self.hyde_llm is not None:
            hyde = HyDEQueryTransform(llm=self.hyde_llm, include_original=True)
            retriever = TransformQueryEngine(retriever, query_transform=hyde)  # type: ignore
        return [
            QueryEngineTool(
                query_engine=retriever,
                metadata=ToolMetadata(
                    name=self.dataset_name.replace("/", "_"),
                    description=self.dataset_description,
                ),
            )
        ]

    @property
    def query_engine(self) -> BaseQueryEngine:
        synth = get_response_synthesizer(
            llm=self.response_synthesizer_llm,
            use_async=True,
            response_mode=ResponseMode.TREE_SUMMARIZE,
        )

        return SubQuestionQueryEngine.from_defaults(
            query_engine_tools=self.tools,
            llm=self.subquestion_engine_llm,
            verbose=self.verbose,
            response_synthesizer=synth,
            use_async=False,
        )


@dataclass(kw_only=True)
class AgenticRAGFlow(RAGFlow):
    """Base agentic RAG flow with common fields and methods."""

    dataset_name: str
    dataset_description: str
    name: str = "Agentic RAG Flow"

    def __repr__(self):
        return f"{self.name}: {self.params}"

    @property
    def tools(self) -> T.List[BaseTool]:
        return [
            QueryEngineTool(
                query_engine=self.query_engine,
                metadata=ToolMetadata(
                    name=self.dataset_name.replace("/", "_"),
                    description=self.dataset_description,
                ),
            )
        ]

    @property
    def agent(self) -> AgentRunner:
        raise NotImplementedError()

    @dispatcher.span
    def _generate(
        self,
        query: str,
        invocation_id: str,
    ) -> T.Tuple[CompletionResponse, float]:
        start_time = time.perf_counter()
        query = self.set_thinking(query)
        response: AgentChatResponse = self.agent.chat(query)
        try:
            completion_response = CompletionResponse(text=response.response)
        except TypeError:
            logger.error("Incorrect response from an agent: %s", response)
            raise
        except Exception:
            logger.error(
                f"Agent {type(super()).__name__} failed to generate a response using parameters: {self.params}."
            )
            raise
        duration = time.perf_counter() - start_time
        return completion_response, duration

    @dispatcher.span
    async def _agenerate(
        self,
        query: str,
        invocation_id: str,
    ) -> T.Tuple[CompletionResponse, float]:
        start_time = time.perf_counter()
        query = self.set_thinking(query)
        response: AgentChatResponse = await self.agent.achat(query)
        try:
            completion_response = CompletionResponse(text=response.response)
        except TypeError:
            logger.error("Incorrect response from an agent: %s", response)
            raise
        except Exception:
            logger.error(
                f"Agent {type(super()).__name__} failed to generate a response using parameters: {self.params}."
            )
            raise
        duration = time.perf_counter() - start_time
        return completion_response, duration


@dataclass(kw_only=True)
class ReActAgentFlow(AgenticRAGFlow):
    react_prompt: str = get_react_template()
    subquestion_engine_llm: LLM
    max_iterations: int = 10
    name: str = "ReAct Agentic RAG Flow"
    subquestion_response_synthesizer_llm: LLM

    def __repr__(self):
        return f"{self.name}: {self.params}"

    @property
    def agent(self) -> AgentRunner:
        synth = get_response_synthesizer(
            llm=self.subquestion_response_synthesizer_llm,
            use_async=True,
            response_mode=ResponseMode.TREE_SUMMARIZE,
            text_qa_template=self.prompt_template,
        )
        sub_question_engine = SubQuestionQueryEngine.from_defaults(
            query_engine_tools=self.tools,  # type: ignore
            llm=self.subquestion_engine_llm,
            verbose=self.verbose,
            response_synthesizer=synth,
            use_async=False,
        )

        tools = [
            QueryEngineTool(
                query_engine=sub_question_engine,
                metadata=ToolMetadata(
                    name=self.dataset_name.replace("/", "_"),
                    description=self.dataset_description,
                ),
            )
        ]

        formatter = ReActChatFormatter.from_defaults(system_header=self.react_prompt)
        return ReActAgent.from_tools(
            tools,  # type: ignore
            llm=self.response_synthesizer_llm,
            max_iterations=self.max_iterations,
            react_chat_formatter=formatter,
            default_tool_choice="any",
            verbose=self.verbose,
        )


@dataclass(kw_only=True)
class CritiqueAgentFlow(AgenticRAGFlow):
    subquestion_engine_llm: LLM
    subquestion_response_synthesizer_llm: LLM
    critique_agent_llm: FunctionCallingLLM
    reflection_agent_llm: FunctionCallingLLM
    max_iterations: int = 10
    critique_template = get_critique_template()
    stopping_condition: StoppingCallable = lambda critique_str: "PASS" in critique_str
    name: str = "Critique Agentic RAG Flow"

    def __repr__(self):
        return f"{self.name}: {self.params}"

    @property
    def agent(self) -> AgentRunner:
        assert isinstance(self.response_synthesizer_llm, FunctionCallingLLM), (
            f"CritiqueAgentFlow requires FunctionCallingLLM. Got {type(self.response_synthesizer_llm)=}"
        )
        main_worker = FunctionCallingAgentWorker.from_tools(
            tools=self.tools, llm=self.response_synthesizer_llm, verbose=self.verbose
        )
        synth = get_response_synthesizer(
            llm=self.subquestion_response_synthesizer_llm,
            use_async=True,
            response_mode=ResponseMode.TREE_SUMMARIZE,
            text_qa_template=self.prompt_template,
        )
        sub_question_engine = SubQuestionQueryEngine.from_defaults(
            query_engine_tools=self.tools,  # type: ignore
            llm=self.subquestion_engine_llm,
            verbose=self.verbose,
            response_synthesizer=synth,
            use_async=False,
        )

        tools = [
            QueryEngineTool(
                query_engine=sub_question_engine,
                metadata=ToolMetadata(
                    name=self.dataset_name.replace("/", "_"),
                    description=self.dataset_description,
                ),
            )
        ]

        critique_agent_worker = FunctionCallingAgentWorker.from_tools(
            tools=tools,  # type: ignore
            llm=self.critique_agent_llm,
            verbose=self.verbose,
        )

        critique_worker = ToolInteractiveReflectionAgentWorker.from_defaults(
            max_iterations=self.max_iterations,
            critique_agent_worker=critique_agent_worker,
            critique_template=self.critique_template,
            stopping_callable=self.stopping_condition,
            correction_llm=self.reflection_agent_llm,
            verbose=self.verbose,
        )

        introspective_agent_worker = IntrospectiveAgentWorker.from_defaults(
            reflective_agent_worker=critique_worker,
            main_agent_worker=main_worker,
            verbose=self.verbose,
        )

        chat_history = [
            ChatMessage(
                content="You are an assistant that fixes incorrect user-provided answers.",
                role=MessageRole.SYSTEM,
            )
        ]

        return introspective_agent_worker.as_agent(
            verbose=self.verbose,
            chat_history=chat_history,
            default_tool_choice="any",
        )


@dataclass(kw_only=True)
class LATSAgentFlow(AgenticRAGFlow):
    dataset_name: str
    dataset_description: str
    name: str = "LATS Agent Flow"
    num_expansions: int = 2
    max_rollouts: int = 3

    def __repr__(self):
        return f"{self.name}: {self.params}"

    @property
    def tools(self) -> T.List[BaseTool]:
        return [
            QueryEngineTool(
                query_engine=self.query_engine,
                metadata=ToolMetadata(
                    name=self.dataset_name.replace("/", "_"),
                    description=self.dataset_description,
                ),
            )
        ]

    @property
    def agent(self) -> AgentRunner:
        agent_worker = LATSAgentWorker.from_tools(
            self.tools,
            llm=self.response_synthesizer_llm,
            num_expansions=self.num_expansions,
            max_rollouts=self.max_rollouts,
            verbose=self.verbose,
        )
        return agent_worker.as_agent(
            default_tool_choice="any",
        )


@dataclass(kw_only=True)
class CoAAgentFlow(AgenticRAGFlow):
    name: str = "CoA Agent Flow"
    enable_calculator: bool = False

    @property
    def tools(self) -> T.List[BaseTool]:
        tools: T.List[BaseTool] = [
            QueryEngineTool(
                query_engine=self.query_engine,
                metadata=ToolMetadata(
                    name=self.dataset_name.replace("/", "_"),
                    description=self.dataset_description,
                ),
            ),
        ]

        if self.enable_calculator:

            def add(a: int, b: int):
                """Add two numbers together"""
                return a + b

            def subtract(a: int, b: int):
                """Subtract b from a"""
                return a - b

            def multiply(a: int, b: int):
                """Multiply two numbers together"""
                return a * b

            def divide(a: int, b: int):
                """Divide a by b"""
                return a / b

            code_tools = [
                FunctionTool.from_defaults(fn=fn)
                for fn in [add, subtract, multiply, divide]
            ]
            tools += code_tools
        return tools

    @property
    def agent(self) -> AgentRunner:
        pack = CoAAgentPack(
            tools=self.tools,
            llm=self.response_synthesizer_llm,
        )
        return pack.agent


class Flows(Enum):
    GENERATOR_FLOW = Flow
    RAG_FLOW = RAGFlow
    LLAMA_INDEX_REACT_AGENT_FLOW = ReActAgentFlow
    LLAMA_INDEX_CRITIQUE_AGENT_FLOW = CritiqueAgentFlow
    LLAMA_INDEX_SUB_QUESTION_FLOW = SubQuestionRAGFlow
    LLAMA_INDEX_LATS_RAG_AGENT = LATSAgentFlow
    LLAMA_INDEX_COA_RAG_AGENT = CoAAgentFlow
    RETRIEVER_FLOW = RetrieverFlow



import pytest

from syftr.embeddings.timeouts import EmbeddingPreemptiveTimeoutError
from syftr.studies import TimeoutConfig


@pytest.mark.parametrize(
    "total_chunks,total_time,chunks_processed,should_time_out",
    [
        (0, 1000, 10, False),  # Too slow, but shouldn't check
        (1000, 10, 10, False),  # Fast enough to complete
        (1000, 10, 11, False),  # Fast enough to complete
        (1000, 11, 10, True),  # Not fast enough to complete
        (1000, 990, 995, False),  # Almost done
        (1000, 990, 10, True),  # Not enough chunks, but enough time elapsed
        (100000, 5, 101, True),  # Not enough time elapsed, but enough chunks
    ],
)
def test_timeout_logic(
    total_chunks, total_time, chunks_processed, should_time_out, bge_small
):
    bge_small.reset_timeouts(
        timeout_config=TimeoutConfig(
            embedding_timeout_active=True,
            embedding_max_time=1000,
            embedding_min_chunks_to_process=100,
            embedding_min_time_to_process=10,
        )
    )
    bge_small._total_chunks = total_chunks
    bge_small._total_time = total_time
    bge_small._chunks_processed = chunks_processed

    if should_time_out:
        with pytest.raises(EmbeddingPreemptiveTimeoutError):
            bge_small._check_remaining_time()
    else:
        bge_small._check_remaining_time()



from syftr.flows import Flow


def test_token_counts_basic_generator(basic_flow_all_llms):
    response, duration, call_data = basic_flow_all_llms.generate(
        "What is the capital of France?"
    )
    assert "paris" in str(response.text).lower()
    assert duration
    assert len(call_data) >= 1, call_data
    assert sum(call.cost for call in call_data) > 0, call_data
    assert sum(call.input_tokens for call in call_data) > 0, call_data
    assert sum(call.output_tokens for call in call_data) > 0, call_data


def test_token_counts_configurable_models(configured_models):
    name, llm = configured_models
    flow = Flow(response_synthesizer_llm=llm)
    response, duration, call_data = flow.generate("What is the capital of France?")
    assert "paris" in str(response.text).lower()
    assert duration
    assert len(call_data) >= 1, call_data
    assert sum(call.cost for call in call_data) > 0, call_data
    assert sum(call.input_tokens for call in call_data) > 0, call_data
    assert sum(call.output_tokens for call in call_data) > 0, call_data



from llama_index.core.instrumentation import get_dispatcher
from openinference.instrumentation import OITracer, TraceConfig
from openinference.instrumentation.llama_index._handler import EventHandler
from openinference.instrumentation.llama_index.version import __version__
from opentelemetry import trace as trace_api
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry.sdk.trace.export import SimpleSpanProcessor

from syftr.configuration import cfg


def instrument_arize(endpoint: str = cfg.instrumentation.otel_endpoint) -> None:
    """Arize global handler conflicts with our token tracker.

    This accomplishes the same thing.

    Also compatible with OTEL tracing, eg. otel-collector
    """
    tracer_provider = trace_sdk.TracerProvider()
    tracer_provider.add_span_processor(SimpleSpanProcessor(OTLPSpanExporter(endpoint)))

    tracer = OITracer(
        trace_api.get_tracer(__name__, __version__, tracer_provider),
        config=TraceConfig(),
    )

    _event_handler = EventHandler(tracer=tracer)
    _span_handler = _event_handler._span_handler
    dispatcher = get_dispatcher()
    for span_handler in dispatcher.span_handlers:
        if span_handler.__class__.__name__ == "_SpanHandler":
            break
    else:
        dispatcher.add_span_handler(_span_handler)
    for event_handler in dispatcher.event_handlers:
        if event_handler.__class__.__name__ == "EventHandler":
            break
    else:
        dispatcher.add_event_handler(_event_handler)



import argparse

from ray.job_submission import JobStatus

from syftr.configuration import cfg
from syftr.ray.submit import get_client


def cancel_jobs(
    substring: str = "<doesntmatch>",
    prefix: str = "<doesntmatch>",
    suffix: str = "<doesntmatch>",
    remote: bool = True,
):
    """Usage:

    from syftr.scripts.run_benchmarks import cancel_jobs
    cancel_jobs(prefix='bench3', remote=True)
    """
    cfg.ray.local = False if remote else cfg.ray.local

    client = get_client()
    job_details = client.list_jobs()
    jobs_to_stop = [
        job
        for job in job_details
        if job.submission_id is not None
        and (
            job.submission_id.startswith(prefix)
            or job.submission_id.endswith(suffix)
            or substring in job.submission_id
        )
        and job.status not in {JobStatus.STOPPED, JobStatus.SUCCEEDED, JobStatus.FAILED}
    ]

    for job in jobs_to_stop:
        client.stop_job(job.submission_id)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--remote",
        help="Use remote Ray cluster",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--substring",
        type=str,
        help="The substring to match in the job submission ID",
    )
    args = parser.parse_args()
    input_str = args.substring
    cancel_jobs(substring=input_str)



import time
import typing as T
from abc import ABC, abstractmethod

import optuna
import pandas as pd
from optuna import Study, TrialPruned
from optuna.storages import BaseStorage
from paretoset import paretoset

from syftr.configuration import cfg
from syftr.logger import logger
from syftr.optuna_helper import get_completed_trials


class SyftrPruner(ABC):
    @abstractmethod
    def report(
        self,
        step: int,
        **kwargs,
    ):
        pass

    @abstractmethod
    def should_prune(self) -> bool:
        pass


class CostPruner(SyftrPruner):
    _step: int
    _process_time: float

    def __init__(
        self,
        num_warmup_steps: int,
        num_total_steps,
        max_cost: float,
    ):
        self.num_warmup_steps = num_warmup_steps
        self.num_total_steps = num_total_steps
        self.max_cost = max_cost

    def report(
        self,
        step: int,
        total_cost: float = 0.0,
        **kwargs,
    ):
        self._total_cost = total_cost
        self._step = step

    def should_prune(self) -> bool:
        if self._step < self.num_warmup_steps:
            logger.debug("Not pruning because this is a warmup step")
            return False
        return (self._total_cost * self.num_total_steps / self._step) > self.max_cost

    def report_and_raise_on_prune(
        self, step: int, total_cost: float, llm_cost_mean: float
    ):
        self.report(
            step=step,
            total_cost=total_cost,
        )
        if self.should_prune():
            msg = f"Pruned after {step} evaluations because of expected cost-out with mean_cost={llm_cost_mean}"
            logger.warning(msg)
            raise TrialPruned(msg)


class RuntimePruner(SyftrPruner):
    _step: int
    _process_time: float

    def __init__(
        self,
        num_warmup_steps: int,
        num_total_steps,
        eval_timeout: int,
        eval_start: float | None = None,
    ):
        self.num_warmup_steps = num_warmup_steps
        self.num_total_steps = num_total_steps
        self.eval_timeout = eval_timeout
        self.eval_start = eval_start or time.process_time()

    def report(
        self,
        step: int,
        **kwargs,
    ):
        self._process_time = time.process_time()
        self._step = step

    def should_prune(self) -> bool:
        if self._step < self.num_warmup_steps:
            logger.debug("Not pruning because this is a warmup step")
            return False
        return (
            (self._process_time - self.eval_start) * self.num_total_steps / self._step
        ) > self.eval_timeout

    def report_and_raise_on_prune(self, step: int, p80_time: float):
        self.report(step)
        if self.should_prune():
            msg = f"Pruned after {step} evaluations because of expected timeout with p80_time={p80_time}"
            logger.warning(msg)
            raise TrialPruned(msg)


class ParetoPruner(SyftrPruner):
    _obj1: float
    _obj2: float
    _obj1_confidence: float
    _obj2_confidence: float
    _step: int

    def __init__(
        self,
        study_name,
        num_warmup_steps: int,
        success_rate: float | None = None,
        storage: str | BaseStorage | None = None,
    ):
        self.study_name = study_name
        self.num_warmup_steps = num_warmup_steps
        self.directions = ["max", "min"]
        self.num_pareto_points = 1000
        self.success_rate = success_rate
        self.storage = storage if storage else cfg.database.get_optuna_storage()

    def get_pareto_points(self, trials: pd.DataFrame) -> pd.DataFrame:
        pareto_mask = paretoset(trials[["values_0", "values_1"]], sense=self.directions)
        return trials[pareto_mask].copy()

    def point_is_dominated(
        self, pareto_points: pd.DataFrame, point: T.Tuple[float, float]
    ) -> bool:
        point_x = point[0]
        point_y = point[1]
        for pareto_trial in pareto_points[["values_0", "values_1"]].itertuples():
            trial_x = pareto_trial.values_1  # latency
            trial_y = pareto_trial.values_0  # acc
            if trial_x < point_x and trial_y >= point_y:
                return True
            if trial_x <= point_x and trial_y > point_y:
                return True
        return False

    def prune(self, study: Study) -> bool:
        if self._step < self.num_warmup_steps:
            logger.debug("Not pruning because this is a warmup step")
            return False

        df_trials: pd.DataFrame = get_completed_trials(
            study, success_rate=self.success_rate
        )
        if df_trials.empty:
            logger.debug("Not pruning because there are no other trials")
            return False

        point = (
            self._obj2 - self._obj2_confidence,  # latency
            self._obj1 + self._obj1_confidence,  # acc
        )

        pareto_points = self.get_pareto_points(df_trials)

        if self.point_is_dominated(pareto_points, point):
            return True

        return False

    def report(
        self,
        step: int,
        **kwargs,
    ):
        assert "obj1" in kwargs, "obj1 is missing"
        assert "obj2" in kwargs, "obj2 is missing"
        assert "obj1_confidence" in kwargs, "'obj1_confidence' is missing"
        assert "obj2_confidence" in kwargs, "'obj2_confidence' is missing"
        assert step, "step is missing"

        self._step = step
        self._obj1 = kwargs["obj1"]
        self._obj2 = kwargs["obj2"]
        self._obj1_confidence = kwargs["obj1_confidence"]
        self._obj2_confidence = kwargs["obj2_confidence"]

    def should_prune(self) -> bool:
        study = optuna.load_study(study_name=self.study_name, storage=self.storage)
        return self.prune(study)

    def report_and_prune(
        self,
        step: int,
        obj1: float,
        obj2: float,
        obj1_confidence: float,
        obj2_confidence: float,
    ):
        self.report(
            step=step,
            obj1=obj1,
            obj2=obj2,
            obj1_confidence=obj1_confidence,
            obj2_confidence=obj2_confidence,
        )
        if self.should_prune():
            msg = f"Pruned after{step} batch because of lower-than-expected accuracy: {obj1}"
            logger.warning(msg)
            return True
        return False



import math
import typing as T

from syftr.studies import (
    FewShotRetriever,
    Hybrid,
    LLMConfig,
    QueryDecomposition,
    Reranker,
    Retriever,
    SearchSpace,
    Splitter,
    TopK,
)

MINIMAL_LLMS: T.List[str] = ["llmA", "llmB"]


def _get_minimal_splitter() -> Splitter:
    return Splitter(
        chunk_min_exp=12,
        chunk_max_exp=12,
        chunk_overlap_frac_min=0.5,
        chunk_overlap_frac_max=0.75,
        chunk_overlap_frac_step=0.25,
        methods=["html", "recursive"],
    )


def test_splitter_cardinality():
    assert _get_minimal_splitter().get_cardinality() == 4


def _get_minimal_topk() -> TopK:
    return TopK(
        kmin=2,
        kmax=2,
        step=1,
    )


def test_topk_cardinality():
    assert _get_minimal_topk().get_cardinality() == 1


def _get_minmal_llm_config() -> LLMConfig:
    return LLMConfig(
        llm_names=MINIMAL_LLMS,
        llm_temperature_min=0,
        llm_temperature_max=0,
        llm_top_p_min=1.0,
        llm_top_p_max=1.0,
        llm_use_reasoning=[True],
    )


def _get_minimal_qd() -> QueryDecomposition:
    return QueryDecomposition(
        llm_config=_get_minmal_llm_config(),
        num_queries_min=2,
        num_queries_max=4,
        num_queries_step=2,
    )


def test_query_decomposition_cardinality():
    assert _get_minimal_qd().get_cardinality() == 4


def _get_minimal_hybrid() -> Hybrid:
    return Hybrid(
        bm25_weight_min=0.1,
        bm25_weight_max=0.1,
        bm25_weight_step=0.1,
    )


def test_hybrid_cardinality():
    assert _get_minimal_hybrid().get_cardinality() == 1


def _get_minimal_retriever() -> Retriever:
    return Retriever(
        top_k=_get_minimal_topk(),
        methods=["hybrid"],
        embedding_models=["embA"],
        hybrid=_get_minimal_hybrid(),
        query_decomposition=_get_minimal_qd(),
    )


def test_retiever_cardinality():
    r = _get_minimal_retriever()
    assert (
        r.get_cardinality()
        == r.top_k.get_cardinality()
        * len(r.methods)
        * len(r.query_decomposition_enabled)
        * r.hybrid.get_cardinality()
        * r.query_decomposition.get_cardinality()
        * r.fusion.get_cardinality()
    )


def _get_minimal_few_shot_retriever() -> FewShotRetriever:
    return FewShotRetriever(
        top_k=_get_minimal_topk(),
        embedding_models=["embA"],
    )


def _get_minimal_reranker() -> Reranker:
    return Reranker(
        llm_config=_get_minmal_llm_config(),
        top_k=_get_minimal_topk(),
    )


def test_search_space_cardinality():
    # Construct a rag and no_rag search space and check that they match the size
    # of a hybrid rag + no_rag space.
    ss = SearchSpace(
        rag_modes=["rag"],
        template_names=["templateA", "templateB"],
        response_synthesizer_llm_config=LLMConfig(
            llm_names=MINIMAL_LLMS,
            llm_temperature_min=0,
            llm_temperature_max=0,
            llm_top_p_min=1.0,
            llm_top_p_max=1.0,
            llm_use_reasoning=[True],
        ),
        few_shot_retriever=_get_minimal_few_shot_retriever(),
        splitter=_get_minimal_splitter(),
        reranker=_get_minimal_reranker(),
        rag_retriever=_get_minimal_retriever(),
    )
    components = [
        ss.few_shot_retriever,
        ss.rag_retriever,
        ss.splitter,
        ss.reranker,
        ss.hyde,
        ss.additional_context,
    ]
    expected_rag_card = math.prod(
        [component.get_cardinality() for component in components]
    )
    expected_rag_card *= (
        len(ss.template_names)
        * len(MINIMAL_LLMS)
        * len(ss.few_shot_enabled)
        * len(ss.hyde_enabled)
        * len(ss.additional_context_enabled)
        * len(ss.reranker_enabled)
    )
    assert ss.get_cardinality() == expected_rag_card

    ss = SearchSpace(
        rag_modes=["no_rag"],
        template_names=["templateA", "templateB"],
        response_synthesizer_llm_config=_get_minmal_llm_config(),
        few_shot_retriever=_get_minimal_few_shot_retriever(),
        splitter=_get_minimal_splitter(),
        reranker=_get_minimal_reranker(),
        rag_retriever=_get_minimal_retriever(),
    )
    expected_no_rag_card = (
        len(ss.template_names)
        * len(MINIMAL_LLMS)
        * len(ss.few_shot_enabled)
        * len(ss.hyde_enabled)
        * len(ss.additional_context_enabled)
        * len(ss.reranker_enabled)
        * ss.few_shot_retriever.get_cardinality()
    )
    assert ss.get_cardinality() == expected_no_rag_card

    # Now test with both rag and no rag
    ss = SearchSpace(
        rag_modes=["no_rag", "rag"],
        template_names=["templateA", "templateB"],
        response_synthesizer_llm_config=_get_minmal_llm_config(),
        few_shot_retriever=_get_minimal_few_shot_retriever(),
        splitter=_get_minimal_splitter(),
        reranker=_get_minimal_reranker(),
        rag_retriever=_get_minimal_retriever(),
    )
    assert ss.get_cardinality() == (expected_rag_card + expected_no_rag_card)
